"""
Test external subsystem bus API.
"""
from copy import deepcopy
import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from aviary.interface.default_phase_info.height_energy import phase_info as ph_in
from aviary.interface.methods_for_level2 import AviaryProblem
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase


class PreMissionComp(om.ExplicitComponent):

    def setup(self):
        self.add_output('for_climb', np.ones((2, 1)), units='ft')
        self.add_output('for_cruise', np.ones((2, 1)), units='ft')
        self.add_output('for_descent', np.ones((2, 1)), units='ft')

    def compute(self, inputs, outputs):
        outputs['for_climb'] = np.array([[3.1], [1.7]])
        outputs['for_cruise'] = np.array([[1.2], [4.1]])
        outputs['for_descent'] = np.array([[3.], [8.]])


class MissionComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('shape', default=(1, ))

    def setup(self):
        shape = self.options['shape']
        self.add_input('xx', shape=shape, units='ft')
        self.add_output('yy', shape=shape, units='ft')

    def compute(self, inputs, outputs):
        outputs['yy'] = 2.0 * inputs['xx']


class CustomBuilder(SubsystemBuilderBase):

    def build_pre_mission(self, aviary_inputs):
        return PreMissionComp()

    def build_mission(self, num_nodes, aviary_inputs):
        sub_group = om.Group()
        sub_group.add_subsystem('electric', MissionComp(shape=(2, 1)),
                                promotes_inputs=['*'],
                                promotes_outputs=['*'],
                                )
        return sub_group

    def get_bus_variables(self):
        vars_to_connect = {
            "test.for_climb": {
                "mission_name": ['test.xx'],
                "units": 'ft',
                "shape": (2, 1),
                "phases": ['climb']
            },
            "test.for_cruise": {
                "mission_name": ['test.xx'],
                "units": 'ft',
                "shape": (2, 1),
                "phases": ['cruise']
            },
            "test.for_descent": {
                "mission_name": ['test.xx'],
                "units": 'ft',
                "shape": (2, 1),
                "phases": ['descent']
            },
        }

        return vars_to_connect


class TestExternalSubsystemBus(unittest.TestCase):

    def test_external_subsystem_bus(self):
        phase_info = deepcopy(ph_in)
        phase_info['pre_mission']['external_subsystems'] = [CustomBuilder(name='test')]
        phase_info['climb']['external_subsystems'] = [CustomBuilder(name='test')]
        phase_info['cruise']['external_subsystems'] = [CustomBuilder(name='test')]
        phase_info['descent']['external_subsystems'] = [CustomBuilder(name='test')]

        prob = AviaryProblem()

        csv_path = "models/test_aircraft/aircraft_for_bench_FwFm.csv"
        prob.load_inputs(csv_path, phase_info)
        prob.check_and_preprocess_inputs()

        prob.add_pre_mission_systems()
        prob.add_phases()
        prob.add_post_mission_systems()
        prob.link_phases()

        prob.setup()
        prob.set_initial_guesses()

        # Just run once to pass data.
        prob.run_model()

        # Each phase should have a different value passed from pre using the bus.
        assert_near_equal(
            prob.model.get_val('traj.climb.rhs_all.test.yy'),
            2.0 * np.array([[3.1], [1.7]]),
        )
        assert_near_equal(
            prob.model.get_val('traj.cruise.rhs_all.test.yy'),
            2.0 * np.array([[1.2], [4.1]]),
        )
        assert_near_equal(
            prob.model.get_val('traj.descent.rhs_all.test.yy'),
            2.0 * np.array([[3.], [8.]]),
        )


if __name__ == '__main__':
    unittest.main()



from pathlib import Path

from openmdao.utils.reports_system import register_report


def register_custom_reports():
    """
    Registers Aviary reports with openMDAO, so they are automatically generated and
    added to the same reports folder as other default reports
    """
    # TODO top-level aircraft report?
    # TODO mission report?
    # TODO add flag to skip registering reports?

    # register per-subsystem report generation
    register_report(name='subsystems',
                    func=subsystem_report,
                    desc='Generates reports for each subsystem builder in the '
                         'Aviary Problem',
                    class_name='Problem',
                    method='run_model',
                    pre_or_post='post',
                    # **kwargs
                    )


def subsystem_report(prob, **kwargs):
    """
    Loops through all subsystem builders in the AviaryProblem calls their write_report
    method. All generated report files are placed in the ./reports/subsystem_reports folder

    Parameters
    ----------
    prob : AviaryProblem
        The AviaryProblem that will be used to generate this report
    """
    reports_folder = Path(prob.get_reports_dir() / 'subsystems')
    reports_folder.mkdir(exist_ok=True)

    # TODO external subsystems??
    core_subsystems = prob.core_subsystems

    for subsystem in core_subsystems.values():
        subsystem.report(prob, reports_folder, **kwargs)



"""
OpenMDAO System to compute drag based on the methods in FLOPS AERO.
"""
import numpy as np
import openmdao.api as om

from aviary.subsystems.aerodynamics.aero_common import DynamicPressure
from aviary.subsystems.aerodynamics.flops_based.buffet_lift import BuffetLift
from aviary.subsystems.aerodynamics.flops_based.compressibility_drag import \
    CompressibilityDrag
from aviary.subsystems.aerodynamics.flops_based.drag import TotalDrag
from aviary.subsystems.aerodynamics.flops_based.induced_drag import InducedDrag
from aviary.subsystems.aerodynamics.flops_based.lift import LiftEqualsWeight
from aviary.subsystems.aerodynamics.flops_based.lift_dependent_drag import \
    LiftDependentDrag
from aviary.subsystems.aerodynamics.flops_based.mux_component import MuxComponent
from aviary.subsystems.aerodynamics.flops_based.skin_friction import SkinFriction
from aviary.subsystems.aerodynamics.flops_based.skin_friction_drag import \
    SkinFrictionDrag
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class ComputedAeroGroup(om.Group):

    def initialize(self):
        self.options.declare(
            "num_nodes", default=1, types=int,
            desc="Number of nodes along mission segment")
        self.options.declare(
            'gamma', default=1.4,
            desc='Ratio of specific heats for air.')
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        num_nodes = self.options["num_nodes"]
        gamma = self.options['gamma']
        aviary_options: AviaryValues = self.options['aviary_options']

        comp = MuxComponent(aviary_options=aviary_options)
        self.add_subsystem(
            'Mux', comp,
            promotes_inputs=['aircraft:*'],
            promotes_outputs=[
                'wetted_areas', 'fineness_ratios', 'characteristic_lengths',
                'laminar_fractions_upper', 'laminar_fractions_lower'])

        self.add_subsystem(
            'DynamicPressure', DynamicPressure(num_nodes=num_nodes, gamma=gamma),
            promotes_inputs=[Dynamic.Mission.MACH, Dynamic.Mission.STATIC_PRESSURE],
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE])

        comp = LiftEqualsWeight(num_nodes=num_nodes)
        self.add_subsystem(
            name=Dynamic.Mission.LIFT, subsys=comp,
            promotes_inputs=[Aircraft.Wing.AREA, Dynamic.Mission.MASS,
                             Dynamic.Mission.DYNAMIC_PRESSURE],
            promotes_outputs=['cl', Dynamic.Mission.LIFT])

        comp = LiftDependentDrag(num_nodes=num_nodes, gamma=gamma)
        self.add_subsystem(
            'PressureDrag', comp,
            promotes_inputs=[
                Dynamic.Mission.MACH, Dynamic.Mission.LIFT, Dynamic.Mission.STATIC_PRESSURE,
                Mission.Design.MACH,
                Mission.Design.LIFT_COEFFICIENT,
                Aircraft.Wing.AREA,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.THICKNESS_TO_CHORD])

        comp = InducedDrag(
            num_nodes=num_nodes, gamma=gamma, aviary_options=aviary_options)
        self.add_subsystem(
            'InducedDrag', comp,
            promotes_inputs=[
                Dynamic.Mission.MACH, Dynamic.Mission.LIFT, Dynamic.Mission.STATIC_PRESSURE,
                Aircraft.Wing.AREA,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SPAN_EFFICIENCY_FACTOR,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.TAPER_RATIO])

        comp = CompressibilityDrag(num_nodes=num_nodes)
        self.add_subsystem(
            'CompressibilityDrag', comp,
            promotes_inputs=[
                Dynamic.Mission.MACH,
                Mission.Design.MACH,
                Aircraft.Design.BASE_AREA,
                Aircraft.Wing.AREA,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.TAPER_RATIO,
                Aircraft.Wing.THICKNESS_TO_CHORD,
                Aircraft.Fuselage.CROSS_SECTION,
                Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
                Aircraft.Fuselage.LENGTH_TO_DIAMETER])

        comp = SkinFriction(num_nodes=num_nodes, aviary_options=aviary_options)
        self.add_subsystem(
            'SkinFrictionCoef', comp,
            promotes_inputs=[
                Dynamic.Mission.MACH, Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.TEMPERATURE,
                'characteristic_lengths'],
            promotes_outputs=['skin_friction_coeff', 'Re'])

        comp = SkinFrictionDrag(num_nodes=num_nodes, aviary_options=aviary_options)
        self.add_subsystem(
            'SkinFrictionDrag', comp,
            promotes_inputs=[
                'skin_friction_coeff', 'Re', 'fineness_ratios', 'wetted_areas',
                'laminar_fractions_upper', 'laminar_fractions_lower',
                Aircraft.Wing.AREA])

        comp = ComputedDrag(num_nodes=num_nodes)
        self.add_subsystem(
            'Drag', comp,
            promotes_inputs=[
                Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.MACH, Aircraft.Wing.AREA,
                Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR,
                Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR,
                Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR,
                Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR],
            promotes_outputs=['CDI', 'CD0', 'drag_coefficient', Dynamic.Mission.DRAG])

        buf = BuffetLift(num_nodes=num_nodes)
        self.add_subsystem(
            'Buffet', buf,
            promotes_inputs=[
                Dynamic.Mission.MACH,
                Mission.Design.MACH,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                Aircraft.Wing.SWEEP,
                Aircraft.Wing.THICKNESS_TO_CHORD])

        self.connect('PressureDrag.CD', 'Drag.pressure_drag_coeff')
        self.connect('InducedDrag.induced_drag_coeff', 'Drag.induced_drag_coeff')
        self.connect(
            'CompressibilityDrag.compress_drag_coeff', 'Drag.compress_drag_coeff')
        self.connect(
            'SkinFrictionDrag.skin_friction_drag_coeff', 'Drag.skin_friction_drag_coeff')


class ComputedDrag(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        self._setup_drag_coeff(
            'CDI',
            input0='pressure_drag_coeff', input1='induced_drag_coeff', output='CDI',
            desc='lift-dependent drag coefficient,'
            ' including contributions from pressure drag coefficient')

        self._setup_drag_coeff(
            'CD0',
            input0='skin_friction_drag_coeff', input1='compress_drag_coeff',
            output='CD0',
            desc='zero-lift drag coefficient')

        self.add_subsystem(
            Dynamic.Mission.DRAG, TotalDrag(num_nodes=nn),
            promotes_inputs=[
                Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR,
                Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR,
                Aircraft.Wing.AREA,
                Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR,
                Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR,
                'CDI', 'CD0', Dynamic.Mission.MACH, Dynamic.Mission.DYNAMIC_PRESSURE],
            promotes_outputs=['drag_coefficient', Dynamic.Mission.DRAG])

        self.set_input_defaults(Aircraft.Wing.AREA, 1., 'ft**2')

    def _setup_drag_coeff(self, name, input0, input1, output, desc=None):
        nn = self.options["num_nodes"]

        input_args = {'val': np.ones(nn), 'units': 'unitless'}
        output_args = dict(input_args)

        if desc is not None:
            output_args['desc'] = desc

        kwargs = {input0: input_args, input1: input_args, output: output_args}

        subsys = self.add_subsystem(
            name, om.ExecComp(f'{output} = {input0} + {input1}', **kwargs),
            promotes_inputs=[input0, input1], promotes_outputs=[output])
        subsys.declare_coloring(show_summary=False)



import unittest

from openmdao.utils.assert_utils import assert_near_equal

from aviary.interface.default_phase_info.two_dof import default_mission_subsystems
from aviary.mission.gasp_based.idle_descent_estimation import descent_range_and_fuel
from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.variable_info.variables import Aircraft, Dynamic
from aviary.utils.process_input_decks import create_vehicle
from aviary.utils.preprocessors import preprocess_propulsion
import importlib


@unittest.skipUnless(importlib.util.find_spec("pyoptsparse") is not None, "pyoptsparse is not installed")
class IdleDescentTestCase(unittest.TestCase):
    def test_case1(self):
        tol = 1e-5
        input_deck = 'models/large_single_aisle_1/large_single_aisle_1_GwGm.csv'
        aviary_inputs, _ = create_vehicle(input_deck)
        aviary_inputs.set_val('debug_mode', False)
        aviary_inputs.set_val(Aircraft.Engine.SCALED_SLS_THRUST, val=28690, units="lbf")
        aviary_inputs.set_val(Dynamic.Mission.THROTTLE, val=0, units="unitless")
        ode_args = dict(aviary_options=aviary_inputs,
                        core_subsystems=default_mission_subsystems)
        engine = EngineDeck(options=aviary_inputs)
        preprocess_propulsion(aviary_inputs, [engine])

        results = descent_range_and_fuel(ode_args=ode_args)['refined_guess']

        # Values obtained by running idle_descent_estimation
        assert_near_equal(results['distance_flown'], 91.8911599691433, tol)
        assert_near_equal(results['fuel_burned'], 236.73893823639082, tol)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.phases.landing_components import (
    GlideConditionComponent, LandingAltitudeComponent,
    LandingGroundRollComponent)
from aviary.variable_info.variables import Aircraft, Mission


class LandingAltTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", LandingAltitudeComponent(), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Mission.Landing.OBSTACLE_HEIGHT, 50, units="ft")
        self.prob.model.set_input_defaults(
            Mission.Landing.AIRPORT_ALTITUDE, 0, units="ft")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob[Mission.Landing.OBSTACLE_HEIGHT], 50, tol
        )  # not actual GASP value, but intuitively correct

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class GlideTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", GlideConditionComponent(), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            "rho_app", 0.0023737, units="slug/ft**3"
        )  # value from online calculator

        self.prob.model.set_input_defaults(
            Mission.Landing.MAXIMUM_SINK_RATE, 900, units="ft/min")

        self.prob.model.set_input_defaults("mass", 165279, units="lbm")
        self.prob.model.set_input_defaults(Aircraft.Wing.AREA, 1370.3, units="ft**2")
        self.prob.model.set_input_defaults(
            Mission.Landing.GLIDE_TO_STALL_RATIO, 1.3, units="unitless")
        self.prob.model.set_input_defaults("CL_max", 2.9533, units="unitless")
        self.prob.model.set_input_defaults(
            Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR, 1.15, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Mission.Landing.TOUCHDOWN_SINK_RATE, 5, units="ft/s")
        self.prob.model.set_input_defaults(
            Mission.Landing.INITIAL_ALTITUDE, val=50.0, units="ft")
        self.prob.model.set_input_defaults(
            Mission.Landing.BRAKING_DELAY, val=1.0, units="s")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob.get_val(Mission.Landing.INITIAL_VELOCITY,
                              units="kn"), 142.87924381, tol
        )  # note: actual GASP value is: 142.74
        assert_near_equal(
            self.prob.get_val(Mission.Landing.STALL_VELOCITY,
                              units="kn"), 109.90711062, tol
        )  # note: EAS in GASP, although at this altitude they are nearly identical. actual GASP value is 109.73
        assert_near_equal(
            self.prob.get_val("TAS_touchdown", units="kn"), 126.39317722, tol
        )  # note: actual GASP value is: 126.27
        assert_near_equal(
            self.prob.get_val("density_ratio", units="unitless"), 0.99819176, tol
        )  # note: calculated from GASP glide speed values as: .998739
        assert_near_equal(
            self.prob.get_val("wing_loading_land", units="lbf/ft**2"), 120.61519375, tol
        )  # note: actual GASP value is: 120.61
        assert_near_equal(
            self.prob.get_val("theta", units="deg"), 3.56616698, tol
        )  # note: actual GASP value is: 3.57
        assert_near_equal(
            self.prob.get_val("glide_distance", units="ft"), 802.28678384, tol
        )  # note: actual GASP value is: 802
        assert_near_equal(
            self.prob.get_val("tr_distance", units="ft"), 166.6422152, tol
        )  # note: actual GASP value is: 167
        assert_near_equal(
            self.prob.get_val("delay_distance", units="ft"), 213.32765038, tol
        )  # note: actual GASP value is: 213
        assert_near_equal(
            self.prob.get_val("flare_alt", units="ft"), 20.7340346, tol
        )  # note: actual GASP value is: 20.8

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-12)


class GroundRollTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", LandingGroundRollComponent(), promotes=["*"]
        )

        self.prob.model.set_input_defaults("touchdown_CD", val=0.07344)
        self.prob.model.set_input_defaults("touchdown_CL", val=1.18694)
        self.prob.model.set_input_defaults(
            Mission.Landing.STALL_VELOCITY, val=109.73, units="kn"
        )  # note: EAS in GASP, although at this altitude they are nearly identical
        self.prob.model.set_input_defaults("TAS_touchdown", val=126.27, units="kn")
        self.prob.model.set_input_defaults("thrust_idle", val=1276, units="lbf")
        self.prob.model.set_input_defaults(
            "density_ratio", val=0.998739, units="unitless"
        )  # note: calculated from GASP glide speed values
        self.prob.model.set_input_defaults(
            "wing_loading_land", val=120.61, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults("glide_distance", val=802, units="ft")
        self.prob.model.set_input_defaults("tr_distance", val=167, units="ft")
        self.prob.model.set_input_defaults("delay_distance", val=213, units="ft")
        self.prob.model.set_input_defaults("CL_max", 2.9533, units="unitless")
        self.prob.model.set_input_defaults("mass", 165279, units="lbm")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob["ground_roll_distance"], 2406.43116212, tol
        )  # actual GASP value is: 1798
        assert_near_equal(
            self.prob[Mission.Landing.GROUND_DISTANCE], 3588.43116212, tol
        )  # actual GASP value is: 2980
        assert_near_equal(
            self.prob["average_acceleration"], 0.29308129, tol
        )  # actual GASP value is: 0.3932

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=5e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import copy
import unittest
import warnings

import dymos as dm
import openmdao.api as om
from openmdao.core.driver import Driver
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.subsystems.premission import CorePreMission

from aviary.utils.functions import set_aviary_initial_values

from aviary.models.N3CC.N3CC_data import (
    inputs as _inputs, outputs as _outputs,
    landing_trajectory_builder as _landing_trajectory_builder,
    landing_fullstop_user_options as _landing_fullstop_user_options)

from aviary.variable_info.variables import Dynamic
from aviary.interface.default_phase_info.height_energy import default_premission_subsystems
from aviary.utils.preprocessors import preprocess_crewpayload
from aviary.variable_info.variables_in import VariablesIn


@use_tempdirs
class TestFLOPSDetailedLanding(unittest.TestCase):
    # @require_pyoptsparse(optimizer='IPOPT')
    # def bench_test_IPOPT(self):
    #     driver = om.pyOptSparseDriver()

    #     optimizer = 'IPOPT'
    #     driver.options['optimizer'] = optimizer

    #     driver.opt_settings['max_iter'] = 100
    #     driver.opt_settings['tol'] = 1.0E-6
    #     driver.opt_settings['print_level'] = 4
    #     driver.opt_settings['mu_init'] = 1e-5

    #     self._do_run(driver, optimizer)

    @require_pyoptsparse(optimizer='SNOPT')
    def bench_test_SNOPT(self):
        driver = om.pyOptSparseDriver()

        optimizer = 'SNOPT'
        driver.options['optimizer'] = optimizer

        driver.opt_settings['Major iterations limit'] = 50
        driver.opt_settings['Major optimality tolerance'] = 1e-4
        driver.opt_settings['Major feasibility tolerance'] = 1e-6
        driver.opt_settings['iSumm'] = 6

        self._do_run(driver, optimizer)

    def _do_run(self, driver: Driver, optimizer, *args):
        aviary_options = _inputs.deepcopy()

        landing_trajectory_builder = copy.deepcopy(_landing_trajectory_builder)
        landing_fullstop_user_options = _landing_fullstop_user_options.deepcopy()

        landing = om.Problem()
        landing.driver = driver

        driver.add_recorder(om.SqliteRecorder(
            f'FLOPS_detailed_landing_traj_{optimizer}.sql'))

        driver.recording_options['record_derivatives'] = False

        preprocess_crewpayload(aviary_options)

        # Upstream static analysis for aero
        landing.model.add_subsystem(
            'pre_mission',
            CorePreMission(aviary_options=aviary_options,
                           subsystems=default_premission_subsystems),
            promotes_inputs=['aircraft:*', 'mission:*'],
            promotes_outputs=['aircraft:*', 'mission:*'])

        # Instantiate the trajectory and add the phases
        traj = dm.Trajectory()
        landing.model.add_subsystem('traj', traj)

        landing_trajectory_builder.build_trajectory(
            aviary_options=aviary_options, model=landing.model, traj=traj)

        distance_max, units = landing_fullstop_user_options.get_item('distance_max')
        fullstop = landing_trajectory_builder.get_phase('landing_fullstop')

        fullstop.add_objective(Dynamic.Mission.DISTANCE, loc='final',
                               ref=distance_max, units=units)

        landing.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        # suppress warnings:
        # "input variable '...' promoted using '*' was already promoted using 'aircraft:*'
        with warnings.catch_warnings():
            # Set initial default values for all aircraft variables.
            set_aviary_initial_values(landing.model, aviary_options)

            warnings.simplefilter("ignore", om.PromotionWarning)
            landing.setup(check=True)

        # Turn off solver printing so that the SNOPT output is readable.
        landing.set_solver_print(level=0)

        landing_trajectory_builder.apply_initial_guesses(landing, 'traj')

        # run the problem
        dm.run_problem(landing, run_driver=True, simulate=True, make_plots=False)

        # Field length
        # N3CC FLOPS output line 1773
        base = -954.08  # ft
        # N3CC FLOPS output line 1842
        desired = 3409.47  # ft

        actual = landing.model.get_val(
            'traj.landing_fullstop.states:distance', units='ft')[-1]

        assert_near_equal(actual, desired, 0.05)

        # TOUCHDOWN time
        # N3CC FLOPS output line 1773
        base = -4.08  # s
        # N3CC FLOPS output line 1849
        desired = 4.22  # s

        actual = landing.model.get_val(
            'traj.landing_flare.t', units='s')[-1]

        assert_near_equal(actual, desired, 0.10)

        # End of landing time
        # N3CC FLOPS output line 1852
        desired = 24.49  # s

        actual = landing.model.get_val(
            'traj.landing_fullstop.t', units='s')[-1]

        assert_near_equal(actual, desired, 0.05)


if __name__ == '__main__':
    use_SNOPT = True

    z = TestFLOPSDetailedLanding()

    if use_SNOPT:
        z.bench_test_SNOPT()

    else:
        z.bench_test_IPOPT()



'''
Define utilities for calculating takeoff EOMs.
'''
import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_METRIC_FLOPS as grav_metric

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Dynamic, Mission


class StallSpeed(om.ExplicitComponent):
    '''
    Define a component for calculating the minimum speed of an aircraft required to
    produce lift.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

    def setup(self):
        options = self.options

        nn = options['num_nodes']

        self.add_input(
            'mass', val=np.ones(nn), units='kg', desc='current mass of the aircraft')

        self.add_input(
            'density', val=np.ones(nn), units='kg/m**3',
            desc='current atmospheric density')

        self.add_input(
            'area', val=1., units='m**2', desc='surface area contributing to lift')

        self.add_input(
            'lift_coefficient_max', val=1., units='unitless',
            desc='maximum lift coefficient')

        self.add_output(
            'stall_speed', val=np.zeros(nn), units='m/s',
            desc='minimum speed of an aircraft required to produce lift')

    def setup_partials(self):
        options = self.options

        nn = options['num_nodes']
        rows_cols = np.arange(nn)

        self.declare_partials(
            'stall_speed', ['mass', 'density'], rows=rows_cols, cols=rows_cols)

        self.declare_partials('stall_speed', ['area', 'lift_coefficient_max'])

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mass = inputs['mass']
        density = inputs['density']
        area = inputs['area']
        lift_coefficient_max = inputs['lift_coefficient_max']

        weight = mass * grav_metric

        stall_speed = (2. * weight / (density * area * lift_coefficient_max))**0.5
        outputs['stall_speed'] = stall_speed

    def compute_partials(self, inputs, J, discrete_inputs=None):
        mass = inputs['mass']
        density = inputs['density']
        area = inputs['area']
        lift_coefficient_max = inputs['lift_coefficient_max']

        weight = mass * grav_metric

        stall_speed = (2. * weight / (density * area * lift_coefficient_max))**0.5

        J['stall_speed', 'mass'] = \
            grav_metric / (stall_speed * density * area * lift_coefficient_max)

        J['stall_speed', 'density'] = \
            -weight / (stall_speed * density**2 * area * lift_coefficient_max)

        J['stall_speed', 'area'] = \
            -weight / (stall_speed * density * area**2 * lift_coefficient_max)

        J['stall_speed', 'lift_coefficient_max'] = \
            -weight / (stall_speed * density * area * lift_coefficient_max**2)


class TakeoffEOM(om.Group):
    '''
    Define a group for calculating takeoff equations of motion.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'climbing', default=False, types=bool,
            desc='mode of operation (ground roll or flight)'
        )

        options.declare(
            'friction_key',
            desc='current friction coefficient key, '
            'either rolling friction or braking friction')

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        options = self.options

        nn = options['num_nodes']
        climbing = options['climbing']
        friction_key = options['friction_key']
        aviary_options = options['aviary_options']
        mu = aviary_options.get_val(friction_key)

        kwargs = {
            'num_nodes': nn,
            'climbing': climbing
        }

        inputs = [Dynamic.Mission.FLIGHT_PATH_ANGLE, Dynamic.Mission.VELOCITY]
        outputs = [Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.ALTITUDE_RATE]

        self.add_subsystem(
            'distance_rates', DistanceRates(**kwargs),
            promotes_inputs=inputs, promotes_outputs=outputs
        )

        kwargs = {
            'num_nodes': nn,
            'climbing': climbing,
            'friction_coefficient': mu,
            'aviary_options': aviary_options
        }

        self.add_subsystem(
            'sum_forces',
            SumForces(**kwargs),
            promotes_inputs=['*'],
            promotes_outputs=['forces_horizontal', 'forces_vertical'])

        self.add_subsystem(
            'accelerations', Accelerations(num_nodes=nn),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'velocity_rate', VelocityRate(num_nodes=nn),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'flight_path_angle_rate', FlightPathAngleRate(num_nodes=nn),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        self.add_subsystem(
            'climb_gradient_forces',
            ClimbGradientForces(num_nodes=nn, aviary_options=aviary_options),
            promotes=['*'])


class DistanceRates(om.ExplicitComponent):
    '''
    Define a component for calculating takeoff horizontal and vertical velocity
    components.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'climbing', default=False, types=bool,
            desc='mode of operation (ground roll or flight)'
        )

    def setup(self):
        options = self.options

        nn = options['num_nodes']

        add_aviary_input(self, Dynamic.Mission.FLIGHT_PATH_ANGLE,
                         val=np.zeros(nn), units='rad')
        add_aviary_input(self, Dynamic.Mission.VELOCITY, val=np.zeros(nn), units='m/s')

        add_aviary_output(self, Dynamic.Mission.DISTANCE_RATE,
                          val=np.zeros(nn), units='m/s')
        add_aviary_output(self, Dynamic.Mission.ALTITUDE_RATE,
                          val=np.zeros(nn), units='m/s')

    def setup_partials(self):
        options = self.options

        nn = options['num_nodes']
        rows_cols = np.arange(nn)
        climbing = options['climbing']

        if climbing:
            self.declare_partials('*', '*', rows=rows_cols, cols=rows_cols)

        else:
            self.declare_partials(
                Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.FLIGHT_PATH_ANGLE, dependent=False)

            self.declare_partials(
                Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.VELOCITY, val=np.identity(nn))

            self.declare_partials(Dynamic.Mission.ALTITUDE_RATE, '*', dependent=False)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        velocity = inputs[Dynamic.Mission.VELOCITY]

        if self.options['climbing']:
            flight_path_angle = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]

            cgam = np.cos(flight_path_angle)
            range_rate = cgam * velocity

            sgam = np.sin(flight_path_angle)
            altitude_rate = sgam * velocity

            outputs[Dynamic.Mission.ALTITUDE_RATE] = altitude_rate

        else:
            range_rate = velocity

        outputs[Dynamic.Mission.DISTANCE_RATE] = range_rate

    def compute_partials(self, inputs, J, discrete_inputs=None):
        if self.options['climbing']:
            flight_path_angle = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
            velocity = inputs[Dynamic.Mission.VELOCITY]

            cgam = np.cos(flight_path_angle)
            sgam = np.sin(flight_path_angle)

            J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.FLIGHT_PATH_ANGLE] = -sgam * velocity
            J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.VELOCITY] = cgam

            J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.FLIGHT_PATH_ANGLE] = cgam * velocity
            J[Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.VELOCITY] = sgam


class Accelerations(om.ExplicitComponent):
    '''
    Define a component for calculating horizontal and vertical accelerations from forces.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, Dynamic.Mission.MASS, val=np.ones(nn), units='kg')

        self.add_input(
            'forces_horizontal', val=np.zeros(nn), units='N',
            desc='current sum of forces in the horizontal direction')

        self.add_input(
            'forces_vertical', val=np.zeros(nn), units='N',
            desc='current sum of forces in the vertical direction')

        self.add_output(
            'acceleration_horizontal', val=np.zeros(nn),
            desc='current horizontal acceleration', units='m/s**2')

        self.add_output(
            'acceleration_vertical', val=np.zeros(nn),
            desc='current vertical acceleration', units='m/s**2')

    def setup_partials(self):
        nn = self.options['num_nodes']

        rows_cols = np.arange(nn)

        self.declare_partials(
            'acceleration_horizontal', Dynamic.Mission.MASS, rows=rows_cols, cols=rows_cols)

        self.declare_partials(
            'acceleration_vertical', Dynamic.Mission.MASS, rows=rows_cols, cols=rows_cols)

        self.declare_partials(
            'acceleration_horizontal', 'forces_horizontal', rows=rows_cols,
            cols=rows_cols)

        self.declare_partials(
            'acceleration_vertical', 'forces_horizontal', dependent=False)

        self.declare_partials(
            'acceleration_horizontal', 'forces_vertical', dependent=False)

        self.declare_partials(
            'acceleration_vertical', 'forces_vertical', rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mass = inputs[Dynamic.Mission.MASS]
        f_h = inputs['forces_horizontal']
        f_v = inputs['forces_vertical']

        a_h = f_h / mass
        outputs['acceleration_horizontal'] = a_h

        a_v = f_v / mass
        outputs['acceleration_vertical'] = a_v

    def compute_partials(self, inputs, J, discrete_inputs=None):
        mass = inputs[Dynamic.Mission.MASS]
        f_h = inputs['forces_horizontal']
        f_v = inputs['forces_vertical']

        m2 = mass * mass

        J['acceleration_horizontal', Dynamic.Mission.MASS] = -f_h / m2
        J['acceleration_vertical', Dynamic.Mission.MASS] = -f_v / m2

        J['acceleration_horizontal', 'forces_horizontal'] = 1. / mass

        J['acceleration_vertical', 'forces_vertical'] = 1. / mass


class VelocityRate(om.ExplicitComponent):
    '''
    Define a component for calculating total acceleration.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(
            'acceleration_horizontal', val=np.zeros(nn),
            desc='current horizontal acceleration', units='m/s**2')

        self.add_input(
            'acceleration_vertical', val=np.zeros(nn),
            desc='current vertical acceleration', units='m/s**2')

        add_aviary_input(self, Dynamic.Mission.DISTANCE_RATE,
                         val=np.zeros(nn), units='m/s')
        add_aviary_input(self, Dynamic.Mission.ALTITUDE_RATE,
                         val=np.zeros(nn), units='m/s')

        add_aviary_output(self, Dynamic.Mission.VELOCITY_RATE,
                          val=np.ones(nn), units='m/s**2')

        rows_cols = np.arange(nn)

        self.declare_partials('*', '*', rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        a_h = inputs['acceleration_horizontal']
        a_v = inputs['acceleration_vertical']
        v_h = inputs[Dynamic.Mission.DISTANCE_RATE]
        v_v = inputs[Dynamic.Mission.ALTITUDE_RATE]

        v_mag = np.sqrt(v_h**2 + v_v**2)
        outputs[Dynamic.Mission.VELOCITY_RATE] = (a_h * v_h + a_v * v_v) / v_mag

    def compute_partials(self, inputs, J, discrete_inputs=None):
        a_h = inputs['acceleration_horizontal']
        a_v = inputs['acceleration_vertical']
        v_h = inputs[Dynamic.Mission.DISTANCE_RATE]
        v_v = inputs[Dynamic.Mission.ALTITUDE_RATE]

        num = (a_h * v_h + a_v * v_v)
        fact = v_h**2 + v_v**2
        den = np.sqrt(fact)

        J[Dynamic.Mission.VELOCITY_RATE, 'acceleration_horizontal'] = v_h / den
        J[Dynamic.Mission.VELOCITY_RATE, 'acceleration_vertical'] = v_v / den

        J[Dynamic.Mission.VELOCITY_RATE,
          Dynamic.Mission.DISTANCE_RATE] = a_h / den - 0.5 * num / fact**(3/2) * 2.0 * v_h

        J[Dynamic.Mission.VELOCITY_RATE,
          Dynamic.Mission.ALTITUDE_RATE] = a_v / den - 0.5 * num / fact**(3/2) * 2.0 * v_v


class FlightPathAngleRate(om.ExplicitComponent):
    '''
    Define a component for calculating flight path angle change rate.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

    def setup(self):
        nn = self.options['num_nodes']

        add_aviary_input(self, Dynamic.Mission.DISTANCE_RATE,
                         val=np.zeros(nn), units='m/s')
        add_aviary_input(self, Dynamic.Mission.ALTITUDE_RATE,
                         val=np.zeros(nn), units='m/s')

        self.add_input(
            'acceleration_horizontal', val=np.zeros(nn),
            desc='current horizontal acceleration',
            units='m/s**2'
        )

        self.add_input(
            'acceleration_vertical', val=np.zeros(nn),
            desc='current vertical acceleration',
            units='m/s**2'
        )

        add_aviary_output(
            self, Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, val=np.zeros(nn), units='rad/s')

        rows_cols = np.arange(nn)

        self.declare_partials('*', '*', rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        v_h = inputs[Dynamic.Mission.DISTANCE_RATE]
        v_v = inputs[Dynamic.Mission.ALTITUDE_RATE]
        a_h = inputs['acceleration_horizontal']
        a_v = inputs['acceleration_vertical']

        x = (a_v * v_h - a_h * v_v) / (v_h**2 + v_v**2)

        outputs[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE] = x

    def compute_partials(self, inputs, J, discrete_inputs=None):
        v_h = inputs[Dynamic.Mission.DISTANCE_RATE]
        v_v = inputs[Dynamic.Mission.ALTITUDE_RATE]
        a_h = inputs['acceleration_horizontal']
        a_v = inputs['acceleration_vertical']

        num = (a_v * v_h - a_h * v_v)
        den = (v_h**2 + v_v**2)

        df_dvh = a_v / den - num / den**2 * 2.0 * v_h

        df_dvv = -a_h / den - num / den**2 * 2.0 * v_v

        df_dah = -v_v / den

        df_dav = v_h / den

        J[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, Dynamic.Mission.DISTANCE_RATE] = df_dvh
        J[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, Dynamic.Mission.ALTITUDE_RATE] = df_dvv
        J[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, 'acceleration_horizontal'] = df_dah
        J[Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, 'acceleration_vertical'] = df_dav


class SumForces(om.ExplicitComponent):
    '''
    Define a component for calculating the separate sums for both the horizontal and
    vertical forces.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'climbing', default=False, types=bool,
            desc='mode of operation (ground roll or flight)'
        )

        options.declare(
            'friction_coefficient', default=0.025,
            desc='current friction coefficient, either rolling friction or braking'
            ' friction')

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        options = self.options

        nn = options['num_nodes']

        add_aviary_input(self, Dynamic.Mission.MASS, val=np.ones(nn), units='kg')
        add_aviary_input(self, Dynamic.Mission.LIFT, val=np.ones(nn), units='N')
        add_aviary_input(self, Dynamic.Mission.THRUST_TOTAL, val=np.ones(nn), units='N')
        add_aviary_input(self, Dynamic.Mission.DRAG, val=np.ones(nn), units='N')

        self.add_input('angle_of_attack', val=np.zeros(nn), units='rad')

        add_aviary_input(self, Dynamic.Mission.FLIGHT_PATH_ANGLE,
                         val=np.zeros(nn), units='rad')

        self.add_output(
            'forces_horizontal', val=np.zeros(nn), units='N',
            desc='current sum of forces in the horizontal direction')

        self.add_output(
            'forces_vertical', val=np.zeros(nn), units='N',
            desc='current sum of forces in the vertical direction')

    def setup_partials(self):
        options = self.options

        nn = options['num_nodes']
        climbing = options['climbing']

        rows_cols = np.arange(nn)

        if climbing:
            self.declare_partials('forces_horizontal',
                                  Dynamic.Mission.MASS, dependent=False)

            self.declare_partials(
                'forces_vertical', Dynamic.Mission.MASS, val=-grav_metric, rows=rows_cols,
                cols=rows_cols)

            wrt = [
                Dynamic.Mission.THRUST_TOTAL, Dynamic.Mission.LIFT, Dynamic.Mission.DRAG, 'angle_of_attack',
                Dynamic.Mission.FLIGHT_PATH_ANGLE]

            self.declare_partials('*', wrt, rows=rows_cols, cols=rows_cols)

        else:
            aviary_options: AviaryValues = options['aviary_options']

            mu = options['friction_coefficient']
            val = -grav_metric * mu

            self.declare_partials(
                'forces_horizontal', Dynamic.Mission.MASS, val=val, rows=rows_cols,
                cols=rows_cols)

            self.declare_partials(
                'forces_horizontal', Dynamic.Mission.LIFT, val=mu, rows=rows_cols,
                cols=rows_cols)

            t_inc = aviary_options.get_val(Mission.Takeoff.THRUST_INCIDENCE, 'rad')
            val = np.cos(t_inc) + np.sin(t_inc) * mu

            self.declare_partials(
                'forces_horizontal', Dynamic.Mission.THRUST_TOTAL, val=val, rows=rows_cols,
                cols=rows_cols)

            self.declare_partials(
                'forces_horizontal', Dynamic.Mission.DRAG, val=-1., rows=rows_cols,
                cols=rows_cols)

            self.declare_partials(
                'forces_horizontal', ['angle_of_attack',
                                      Dynamic.Mission.FLIGHT_PATH_ANGLE],
                dependent=False)

            self.declare_partials('forces_vertical', ['*'], dependent=False)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        options = self.options

        climbing = options['climbing']
        aviary_options: AviaryValues = options['aviary_options']

        t_inc = aviary_options.get_val(Mission.Takeoff.THRUST_INCIDENCE, 'rad')

        mass = inputs[Dynamic.Mission.MASS]
        lift = inputs[Dynamic.Mission.LIFT]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]

        weight = mass * grav_metric

        if climbing:
            # NOTE using FLOPS ROTCL
            #    - section: "COMPUTE TRAJECTORY FROM LIFTOFF UNTIL OBSTACLE HEIGHT IS
            #      REACHED"
            #    - variables: FORCH, FORCV
            alpha0 = \
                aviary_options.get_val(Mission.Takeoff.ANGLE_OF_ATTACK_RUNWAY, 'rad')

            alpha = inputs['angle_of_attack']
            gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]

            angle = alpha - alpha0 + t_inc + gamma

            t_h = thrust * np.cos(angle)
            t_v = thrust * np.sin(angle)

            c_gamma = np.cos(gamma)
            s_gamma = np.sin(gamma)

            f_h = t_h - drag * c_gamma - lift * s_gamma

            f_v = t_v - drag * s_gamma + lift * c_gamma - weight

            outputs['forces_vertical'] = f_v

        else:
            # NOTE using FLOPS GRRUN, which neglects angle of attack
            #    - FLOPS ROTCL applies angle of attack to thrust incidence only; angle of
            #      attack is not applied to any other force
            #    - variables: ACC
            mu = options['friction_coefficient']

            t_h = thrust * np.cos(t_inc)
            t_v = thrust * np.sin(t_inc)

            f_h = t_h - drag - (weight - (lift + t_v)) * mu

        outputs['forces_horizontal'] = f_h

    def compute_partials(self, inputs, J, discrete_inputs=None):
        options = self.options

        climbing = options['climbing']

        if not climbing:
            # see setup_partials()
            return

        aviary_options: AviaryValues = options['aviary_options']

        alpha0 = aviary_options.get_val(Mission.Takeoff.ANGLE_OF_ATTACK_RUNWAY, 'rad')
        t_inc = aviary_options.get_val(Mission.Takeoff.THRUST_INCIDENCE, 'rad')

        lift = inputs[Dynamic.Mission.LIFT]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]

        alpha = inputs['angle_of_attack']
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]

        angle = alpha - alpha0 + t_inc + gamma

        c_angle = np.cos(angle)
        s_angle = np.sin(angle)

        c_gamma = np.cos(gamma)
        s_gamma = np.sin(gamma)

        J['forces_horizontal', Dynamic.Mission.THRUST_TOTAL] = c_angle
        J['forces_vertical', Dynamic.Mission.THRUST_TOTAL] = s_angle

        J['forces_horizontal', Dynamic.Mission.LIFT] = -s_gamma
        J['forces_vertical', Dynamic.Mission.LIFT] = c_gamma

        J['forces_horizontal', Dynamic.Mission.DRAG] = -c_gamma
        J['forces_vertical', Dynamic.Mission.DRAG] = -s_gamma

        J['forces_horizontal', 'angle_of_attack'] = -thrust * s_angle
        J['forces_vertical', 'angle_of_attack'] = thrust * c_angle

        J['forces_horizontal', Dynamic.Mission.FLIGHT_PATH_ANGLE] = \
            -thrust * s_angle + drag * s_gamma - lift * c_gamma

        J['forces_vertical', Dynamic.Mission.FLIGHT_PATH_ANGLE] = \
            thrust * c_angle - drag * c_gamma - lift * s_gamma


class ClimbGradientForces(om.ExplicitComponent):
    '''
    Define a component for calculating residual forces for evaluation of climb gradient
    criteria.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        options = self.options

        nn = options['num_nodes']

        add_aviary_input(self, Dynamic.Mission.MASS, val=np.ones(nn), units='kg')
        add_aviary_input(self, Dynamic.Mission.LIFT, val=np.ones(nn), units='N')
        add_aviary_input(self, Dynamic.Mission.THRUST_TOTAL, val=np.ones(nn), units='N')
        add_aviary_input(self, Dynamic.Mission.DRAG, val=np.ones(nn), units='N')

        self.add_input('angle_of_attack', val=np.zeros(nn), units='rad')

        add_aviary_input(self, Dynamic.Mission.FLIGHT_PATH_ANGLE,
                         val=np.zeros(nn), units='rad')

        self.add_output(
            'climb_gradient_forces_horizontal', val=np.zeros(nn), units='N',
            desc='current sum of forces in the horizontal direction; checking for excess'
            ' thrust')

        self.add_output(
            'climb_gradient_forces_vertical', val=np.zeros(nn), units='N',
            desc='current sum of forces in the vertical direction; checking for net zero'
            ' vertical force')

    def setup_partials(self):
        options = self.options

        nn = options['num_nodes']

        rows_cols = np.arange(nn)

        self.declare_partials(
            '*',
            [
                Dynamic.Mission.MASS, Dynamic.Mission.THRUST_TOTAL, 'angle_of_attack',
                Dynamic.Mission.FLIGHT_PATH_ANGLE],
            rows=rows_cols, cols=rows_cols)

        self.declare_partials(
            'climb_gradient_forces_horizontal', Dynamic.Mission.DRAG, val=-1.,
            rows=rows_cols, cols=rows_cols)

        self.declare_partials(
            'climb_gradient_forces_vertical', Dynamic.Mission.DRAG, dependent=False)

        self.declare_partials(
            'climb_gradient_forces_horizontal', Dynamic.Mission.LIFT, dependent=False)

        self.declare_partials(
            'climb_gradient_forces_vertical', Dynamic.Mission.LIFT, val=1.,
            rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        options = self.options

        aviary_options: AviaryValues = options['aviary_options']

        alpha0 = aviary_options.get_val(Mission.Takeoff.ANGLE_OF_ATTACK_RUNWAY, 'rad')
        t_inc = aviary_options.get_val(Mission.Takeoff.THRUST_INCIDENCE, 'rad')

        mass = inputs[Dynamic.Mission.MASS]
        lift = inputs[Dynamic.Mission.LIFT]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]

        weight = mass * grav_metric

        alpha = inputs['angle_of_attack']
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]

        angle = alpha - alpha0 + t_inc

        c_angle = np.cos(angle)
        s_angle = np.sin(angle)

        c_gamma = np.cos(gamma)
        s_gamma = np.sin(gamma)

        # NOTE using FLOPS CLGRAD
        #    - variables: FORCE2, FORCV
        f_h = -drag - weight * s_gamma + thrust * c_angle
        outputs['climb_gradient_forces_horizontal'] = f_h

        f_v = lift - weight * c_gamma + thrust * s_angle
        outputs['climb_gradient_forces_vertical'] = f_v

    def compute_partials(self, inputs, J, discrete_inputs=None):
        options = self.options

        aviary_options: AviaryValues = options['aviary_options']

        alpha0 = aviary_options.get_val(Mission.Takeoff.ANGLE_OF_ATTACK_RUNWAY, 'rad')
        t_inc = aviary_options.get_val(Mission.Takeoff.THRUST_INCIDENCE, 'rad')

        mass = inputs[Dynamic.Mission.MASS]
        lift = inputs[Dynamic.Mission.LIFT]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]

        weight = mass * grav_metric

        alpha = inputs['angle_of_attack']
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]

        angle = alpha - alpha0 + t_inc

        c_angle = np.cos(angle)
        s_angle = np.sin(angle)

        c_gamma = np.cos(gamma)
        s_gamma = np.sin(gamma)

        f_h_key = 'climb_gradient_forces_horizontal'
        f_v_key = 'climb_gradient_forces_vertical'

        J[f_h_key, Dynamic.Mission.MASS] = -grav_metric * s_gamma
        J[f_v_key, Dynamic.Mission.MASS] = -grav_metric * c_gamma

        J[f_h_key, Dynamic.Mission.THRUST_TOTAL] = c_angle
        J[f_v_key, Dynamic.Mission.THRUST_TOTAL] = s_angle

        J[f_h_key, 'angle_of_attack'] = -thrust * s_angle
        J[f_v_key, 'angle_of_attack'] = thrust * c_angle

        J[f_h_key, Dynamic.Mission.FLIGHT_PATH_ANGLE] = -weight * c_gamma
        J[f_v_key, Dynamic.Mission.FLIGHT_PATH_ANGLE] = weight * s_gamma



import numpy as np

from aviary.variable_info.enums import SpeedType
from aviary.mission.gasp_based.phases.time_integration_phases import SGMGroundroll, \
    SGMRotation, SGMAscentCombined, SGMAccel, SGMClimb, SGMCruise, SGMDescent
from aviary.variable_info.variables import Aircraft, Mission, Dynamic

# defaults for 2DOF based forward in time integeration phases


def create_2dof_based_ascent_phases(
    ode_args,
    cruise_alt=35e3,
    cruise_mach=.8,
):
    groundroll_kwargs = dict(
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    groundroll_vals = {
        # special case
        'attr:VR_value': {'val': 'SGMGroundroll_velocity_trigger', 'units': 'kn'},
    }

    rotation_kwargs = dict(
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    rotation_vals = {}

    ascent_kwargs = dict(
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    ascent_vals = {
        't_init_gear': {'val': 10000, 'units': 's'},
        't_init_flaps': {'val': 10000, 'units': 's'},
        'rotation.start_rotation': {'val': 10000, 'units': 's'},  # special case
        # special case
        'attr:fuselage_angle_max': {'val': Aircraft.Design.MAX_FUSELAGE_PITCH_ANGLE, 'units': 'deg'},
    }

    accel_kwargs = dict(
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    accel_vals = {}

    climb1_kwargs = dict(
        input_speed_type=SpeedType.EAS,
        input_speed_units='kn',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    climb1_vals = {
        'alt_trigger': {'val': 10000, 'units': 'ft'},
        'EAS': {'val': 250, 'units': climb1_kwargs['input_speed_units']},
        'speed_trigger': {'val': cruise_mach, 'units': None},
    }

    climb2_kwargs = dict(
        input_speed_type=SpeedType.EAS,
        input_speed_units='kn',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    climb2_vals = {
        'alt_trigger': {'val': cruise_alt, 'units': 'ft'},
        'EAS': {'val': 270, 'units': climb2_kwargs['input_speed_units']},
        'speed_trigger': {'val': cruise_mach, 'units': None},
    }

    climb3_kwargs = dict(
        input_speed_type=SpeedType.MACH,
        input_speed_units='unitless',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    climb3_vals = {
        'alt_trigger': {'val': cruise_alt, 'units': 'ft'},
        'mach': {'val': cruise_mach, 'units': climb3_kwargs['input_speed_units']},
        'speed_trigger': {'val': 0, 'units': None},
    }

    phases = {
        'groundroll': {
            'ode': SGMGroundroll(**groundroll_kwargs),
            'vals_to_set': groundroll_vals,
        },
        'rotation': {
            'ode': SGMRotation(**rotation_kwargs),
            'vals_to_set': rotation_vals,
        },
        'ascent': {
            'ode': SGMAscentCombined(**ascent_kwargs),
            'vals_to_set': ascent_vals,
        },
        'accel': {
            'ode': SGMAccel(**accel_kwargs),
            'vals_to_set': accel_vals,
        },
        'climb1': {
            'ode': SGMClimb(**climb1_kwargs),
            'vals_to_set': climb1_vals,
        },
        'climb2': {
            'ode': SGMClimb(**climb2_kwargs),
            'vals_to_set': climb2_vals,
        },
        'climb3': {
            'ode': SGMClimb(**climb3_kwargs),
            'vals_to_set': climb3_vals,
        },

    }

    return phases


def create_2dof_based_descent_phases(
    ode_args,
    cruise_mach=.8,
):

    descent1_kwargs = dict(
        input_speed_type=SpeedType.MACH,
        input_speed_units="unitless",
        speed_trigger_units='kn',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    descent1_vals = {
        'alt_trigger': {'val': 10000, 'units': 'ft'},
        'mach': {'val': cruise_mach, 'units': descent1_kwargs['input_speed_units']},
        'speed_trigger': {'val': 350, 'units': descent1_kwargs['speed_trigger_units']},
    }

    descent2_kwargs = dict(
        input_speed_type=SpeedType.EAS,
        input_speed_units="kn",
        speed_trigger_units='kn',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    descent2_vals = {
        'alt_trigger': {'val': 10000, 'units': 'ft'},
        'EAS': {'val': 350, 'units': descent2_kwargs['input_speed_units']},
        'speed_trigger': {'val': 0, 'units': descent2_kwargs['speed_trigger_units']},
    }

    descent3_kwargs = dict(
        input_speed_type=SpeedType.EAS,
        input_speed_units="kn",
        speed_trigger_units='kn',
        ode_args=ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    descent3_vals = {
        'alt_trigger': {'val': 1000, 'units': 'ft'},
        'EAS': {'val': 250, 'units': descent3_kwargs['input_speed_units']},
        'speed_trigger': {'val': 0, 'units': descent3_kwargs['speed_trigger_units']},
    }

    phases = {
        'descent1': {
            'ode': SGMDescent(**descent1_kwargs),
            'vals_to_set': descent1_vals,
        },
        'descent2': {
            'ode': SGMDescent(**descent2_kwargs),
            'vals_to_set': descent2_vals,
        },
        'descent3': {
            'ode': SGMDescent(**descent3_kwargs),
            'vals_to_set': descent3_vals,
        },

    }

    return phases



from copy import deepcopy
import os
import unittest

import dymos
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs
from packaging import version

from aviary.api import Mission
from aviary.interface.default_phase_info.solved import phase_info
from aviary.interface.methods_for_level2 import AviaryProblem
from aviary.variable_info.enums import EquationsOfMotion


@unittest.skipIf(version.parse(dymos.__version__) <= version.parse("1.8.0"),
                 "Older version of Dymos treats non-time integration variables differently.")
@use_tempdirs
@require_pyoptsparse(optimizer="SNOPT")
class ProblemPhaseTestCase(unittest.TestCase):
    def bench_test_solved_full_mission(self):
        # Build problem
        local_phase_info = deepcopy(phase_info)
        prob = AviaryProblem()

        input_file = 'models/test_aircraft/aircraft_for_bench_GwGm_solved.csv'
        prob.load_inputs(input_file, local_phase_info)
        prob.aviary_inputs.set_val(Mission.Design.RANGE, 2000.0, units="NM")


# Preprocess inputs
        prob.check_and_preprocess_inputs()

        prob.add_pre_mission_systems()

        prob.add_phases()

        prob.add_post_mission_systems()

        # Link phases and variables
        prob.link_phases()

        prob.add_driver("SNOPT", max_iter=50, use_coloring=True)

        prob.add_design_variables()

        # Load optimization problem formulation
        # Detail which variables the optimizer can control
        prob.add_objective(objective_type="hybrid_objective")

        prob.setup()

        prob.set_initial_guesses()

        prob.run_aviary_problem("dymos_solution.db")


if __name__ == '__main__':
    z = ProblemPhaseTestCase()
    z.bench_test_solved_full_mission()



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.subsystems.aerodynamics.gasp_based.flaps_model.meta_model import \
    MetaModelGroup
from aviary.variable_info.enums import FlapType
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Dynamic

"""
All data is from validation files using standalone flaps model
"""


class MetaModelTestCasePlain(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.FLAP_TYPE, val=FlapType.PLAIN, units='unitless')
        self.prob.model = LuTMMa = MetaModelGroup(aviary_options=options)
        self.prob.setup()

        self.prob.set_val(Aircraft.Wing.FLAP_CHORD_RATIO, 0.3)
        self.prob.set_val("flap_defl_ratio", 40 / 60)
        self.prob.set_val(Aircraft.Wing.ASPECT_RATIO, 10.13)
        self.prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, 0.13966)
        self.prob.set_val("flap_defl", 40.0, units="deg")
        self.prob.set_val(Aircraft.Wing.FLAP_SPAN_RATIO, 0.65)
        self.prob.set_val("slat_defl_ratio", 10 / 20)
        self.prob.set_val(Aircraft.Wing.SLAT_SPAN_RATIO, 0.89761)
        self.prob.set_val("reynolds", 164.78406)
        self.prob.set_val(Dynamic.Mission.MACH, 0.18368)
        self.prob.set_val(Aircraft.Wing.TAPER_RATIO, 0.33)
        self.prob.set_val(Aircraft.Wing.SLAT_SPAN_RATIO, 0.89761)
        self.prob.set_val("body_to_span_ratio", 0.09239)
        self.prob.set_val("chord_to_body_ratio", 0.12679)

    def test_case(self):

        self.prob.run_model()
        tol = 1e-4

        reg_data = 1
        ans = self.prob["VDEL1"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.55667
        ans = self.prob["VDEL2"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.76500
        ans = self.prob["VDEL3"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.05498
        ans = self.prob["fus_lift"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.97217
        ans = self.prob["VLAM1"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.09948
        ans = self.prob["VLAM2"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.97217
        ans = self.prob["VLAM3"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.19742
        ans = self.prob["VLAM4"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1
        ans = self.prob["VLAM5"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.80000
        ans = self.prob["VLAM6"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.73500
        ans = self.prob["VLAM7"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.74000
        ans = self.prob["VLAM10"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.84232
        ans = self.prob["VLAM11"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.03209
        ans = self.prob["VLAM13"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 0.99082
        ans = self.prob["VLAM14"]
        assert_near_equal(ans, reg_data, tol)

        data = self.prob.check_partials(out_stream=None, method="fd")
        assert_check_partials(data, atol=1e-4, rtol=1e-4)


class MetaModelTestCaseSingleSlotted(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.FLAP_TYPE,
                        val=FlapType.SINGLE_SLOTTED, units='unitless')
        self.prob.model = LuTMMb = MetaModelGroup(aviary_options=options)
        self.prob.setup()

        self.prob.set_val(Aircraft.Wing.FLAP_CHORD_RATIO, 0.3)
        self.prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, 0.13966)
        self.prob.set_val("flap_defl", 40.0, units="deg")

    def test_case(self):

        self.prob.run_model()
        tol = 1e-4

        reg_data = 1
        ans = self.prob["VDEL1"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.25725
        ans = self.prob["VLAM4"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1
        ans = self.prob["VLAM5"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.0
        ans = self.prob["VLAM6"]
        assert_near_equal(ans, reg_data, tol)

        data = self.prob.check_partials(out_stream=None, method="fd")
        assert_check_partials(data, atol=1e-4, rtol=1e-4)


class MetaModelTestCaseFowler(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        options = get_option_defaults()
        options.set_val(Aircraft.Wing.FLAP_TYPE, val=FlapType.FOWLER, units='unitless')
        self.prob.model = LuTMMc = MetaModelGroup(aviary_options=options)
        self.prob.setup()

        self.prob.set_val(Aircraft.Wing.FLAP_CHORD_RATIO, 0.3)
        self.prob.set_val("flap_defl", 40.0, units="deg")

    def test_case(self):

        self.prob.run_model()
        tol = 1e-4

        reg_data = 1.0
        ans = self.prob["VLAM5"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.11
        ans = self.prob["VLAM6"]
        assert_near_equal(ans, reg_data, tol)

        data = self.prob.check_partials(out_stream=None, method="fd")
        assert_check_partials(data, atol=1e-4, rtol=1e-4)


if __name__ == "__main__":
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.engine import EngineSize
from aviary.variable_info.variables import Aircraft


class TestEngine(
    unittest.TestCase
):  # this is the GASP test case, input and output values based on large single aisle 1 v3 without bug fix
    def setUp(self):
        self.prob = om.Problem()
        self.prob.model.add_subsystem("engsz", EngineSize(), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Engine.REFERENCE_DIAMETER, 5.8, units="ft")
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALE_FACTOR, val=1.028233
        )
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CORE_DIAMETER_RATIO, 1.25, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.FINENESS, 2, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_large_sinle_aisle_1_defaults(self):
        self.prob.run_model()
        tol = 1e-4
        assert_near_equal(self.prob[Aircraft.Nacelle.AVG_DIAMETER], 7.35163, tol)
        assert_near_equal(self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.70326, tol)
        assert_near_equal(self.prob[Aircraft.Nacelle.SURFACE_AREA], 339.58389, tol)

    def test_partials(self):
        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()



'''
This module is the API for Aviary aircraft analysis code

For users: All built-in Aviary functions, code, and objects 
should be imported from this file.

For developers: All Aviary code which is intended to be
user-facing should be imported to this file.
'''

# TODO: don't rename things here, do it in the entire codebase
# TODO: when documenting methods and classes, make sure to include documentation (printing of docstrings) for everything that's imported in this API
# TODO: remove overload prototype
# TODO: import examples once we settle on those
# TODO: import this in all user-facing files


###################
# General Imports #
###################

from aviary.variable_info.variables import Aircraft, Mission, Dynamic
from aviary.variable_info.options import get_option_defaults, is_option
from aviary.utils.develop_metadata import add_meta_data, update_meta_data
from aviary.variable_info.variable_meta_data import CoreMetaData
from aviary.variable_info.functions import add_aviary_input, add_aviary_output, get_units, override_aviary_vars, setup_trajectory_params
from aviary.utils.merge_hierarchies import merge_hierarchies
from aviary.utils.merge_variable_metadata import merge_meta_data
from aviary.utils.named_values import NamedValues, get_keys, get_items, get_values
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.csv_data_file import read_data_file, write_data_file
from aviary.utils.data_interpolator_builder import build_data_interpolator
from aviary.variable_info.enums import AlphaModes, AnalysisScheme, ProblemType, SpeedType, GASPEngineType, FlapType, EquationsOfMotion, LegacyCode, Verbosity
from aviary.interface.default_phase_info.two_dof import phase_info as default_2DOF_phase_info
from aviary.interface.default_phase_info.two_dof_fiti import create_2dof_based_ascent_phases, create_2dof_based_descent_phases
from aviary.interface.default_phase_info.solved import phase_info as default_solved_phase_info
from aviary.interface.default_phase_info.height_energy import phase_info as default_height_energy_phase_info
from aviary.interface.methods_for_level1 import run_level_1
from aviary.interface.methods_for_level1 import run_aviary
from aviary.interface.methods_for_level2 import AviaryProblem
from aviary.interface.utils.check_phase_info import check_phase_info
from aviary.utils.engine_deck_conversion import EngineDeckConverter
from aviary.utils.Fortran_to_Aviary import create_aviary_deck
from aviary.utils.functions import set_aviary_initial_values, get_path
from aviary.utils.options import list_options
from aviary.constants import GRAV_METRIC_GASP, GRAV_ENGLISH_GASP, GRAV_METRIC_FLOPS, GRAV_ENGLISH_FLOPS, GRAV_ENGLISH_LBM, RHO_SEA_LEVEL_ENGLISH, RHO_SEA_LEVEL_METRIC, MU_TAKEOFF, MU_LANDING, PSLS_PSF, TSLS_DEGR, RADIUS_EARTH_METRIC
from aviary.subsystems.test.subsystem_tester import TestSubsystemBuilderBase, skipIfMissingDependencies
from aviary.interface.default_phase_info.height_energy import default_premission_subsystems, default_mission_subsystems

###################
# Level 3 Imports #
###################

# Miscellaneous
from aviary.interface.methods_for_level2 import PreMissionGroup, PostMissionGroup
from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.subsystems.premission import CorePreMission
from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.utils.preprocessors import preprocess_options, preprocess_propulsion
from aviary.utils.process_input_decks import create_vehicle
from aviary.utils.functions import create_opts2vals, add_opts2vals, Null
from aviary.variable_info.variables_in import VariablesIn
from aviary.utils.preprocessors import preprocess_crewpayload

# ODEs
# TODO: check and see if this works with both sides, or just GASP
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.flops_based.ode.landing_ode import LandingODE as DetailedLandingODE
from aviary.mission.flops_based.ode.landing_ode import FlareODE as DetailedFlareODE
from aviary.mission.flops_based.ode.takeoff_ode import TakeoffODE as DetailedTakeoffODE
from aviary.mission.gasp_based.ode.accel_ode import AccelODE as TwoDOFAccelerationODE
from aviary.mission.gasp_based.ode.ascent_ode import AscentODE as TwoDOFAscentODE
from aviary.mission.gasp_based.ode.breguet_cruise_ode import BreguetCruiseODESolution
from aviary.mission.gasp_based.ode.climb_ode import ClimbODE as TwoDOFClimbODE
from aviary.mission.gasp_based.ode.descent_ode import DescentODE as TwoDOFDescentODE
from aviary.mission.gasp_based.ode.flight_path_ode import FlightPathODE as TwoDOFFlightPathODE
from aviary.mission.gasp_based.ode.groundroll_ode import GroundrollODE as TwoDOFGroundrollODE
from aviary.mission.gasp_based.ode.rotation_ode import RotationODE as TwoDOFRotationODE
from aviary.mission.gasp_based.phases.landing_group import LandingSegment as TwoDOFSimplifiedLanding
from aviary.mission.gasp_based.phases.taxi_group import TaxiSegment as AnalyticTaxi
from aviary.mission.flops_based.phases.simplified_takeoff import TakeoffGroup as HeightEnergySimplifiedTakeoff
from aviary.mission.flops_based.phases.simplified_landing import LandingGroup as HeightEnergySimplifiedLanding


# Phase builders
from aviary.mission.phase_builder_base import PhaseBuilderBase
# note that this is only for simplified right now
from aviary.mission.flops_based.phases.energy_phase import EnergyPhase as HeightEnergyPhaseBuilder
from aviary.mission.flops_based.phases.build_landing import Landing as HeightEnergyLandingPhaseBuilder
# note that this is only for simplified right now
from aviary.mission.flops_based.phases.build_takeoff import Takeoff as HeightEnergyTakeoffPhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingApproachToMicP3 as DetailedLandingApproachToMicP3PhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingMicP3ToObstacle as DetailedLandingMicP3ToObstaclePhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingObstacleToFlare as DetailedLandingObstacleToFlarePhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingFlareToTouchdown as DetailedLandingFlareToTouchdownPhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingTouchdownToNoseDown as DetailedLandingTouchdownToNoseDownPhaseBuilder
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingNoseDownToStop as DetailedLandingNoseDownToStopPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffBrakeReleaseToDecisionSpeed as DetailedTakeoffBrakeReleaseToDecisionSpeedPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffDecisionSpeedToRotate as DetailedTakeoffDecisionSpeedToRotatePhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffDecisionSpeedBrakeDelay as DetailedTakeoffDecisionSpeedBrakeDelayPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffRotateToLiftoff as DetailedTakeoffRotateToLiftoffPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffLiftoffToObstacle as DetailedTakeoffLiftoffToObstaclePhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffObstacleToMicP2 as DetailedTakeoffObstacleToMicP2PhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffMicP2ToEngineCutback as DetailedTakeoffMicP2ToEngineCutbackPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffEngineCutback as DetailedTakeoffEngineCutbackPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffEngineCutbackToMicP1 as DetailedTakeoffEngineCutbackToMicP1PhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffMicP1ToClimb as DetailedTakeoffMicP1ToClimbPhaseBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffBrakeToAbort as DetailedTakeoffBrakeToAbortPhaseBuilder

# Phase builders
from aviary.mission.gasp_based.phases.accel_phase import AccelPhase as TwoDOFAccelerationPhase
from aviary.mission.gasp_based.phases.ascent_phase import AscentPhase as TwoDOFAscentPhase
from aviary.mission.gasp_based.phases.climb_phase import ClimbPhase as TwoDOFClimbPhase
from aviary.mission.gasp_based.phases.descent_phase import DescentPhase as TwoDOFDescentPhase
from aviary.mission.gasp_based.phases.groundroll_phase import GroundrollPhase as TwoDOFGroundrollPhase
from aviary.mission.gasp_based.phases.rotation_phase import RotationPhase as TwoDOFRotationPhase


# Trajectory builders
from aviary.mission.flops_based.phases.detailed_landing_phases import LandingTrajectory as DetailedLandingTrajectoryBuilder
from aviary.mission.flops_based.phases.detailed_takeoff_phases import TakeoffTrajectory as DetailedTakeoffTrajectoryBuilder

# SimuPy
from aviary.mission.gasp_based.ode.time_integration_base_classes import SimuPyProblem
from aviary.mission.gasp_based.phases.time_integration_phases import SGMGroundroll, SGMRotation, SGMAscent, SGMAscentCombined, SGMAccel, SGMClimb, SGMCruise, SGMDescent
from aviary.mission.gasp_based.phases.time_integration_traj import TimeIntegrationTrajBase, FlexibleTraj

# Aerodynamics
from aviary.subsystems.aerodynamics.aerodynamics_builder import AerodynamicsBuilderBase
from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.subsystems.aerodynamics.flops_based.tabular_aero_group import TabularAeroGroup

# Geometry
from aviary.subsystems.geometry.geometry_builder import GeometryBuilderBase
from aviary.subsystems.geometry.geometry_builder import CoreGeometryBuilder

# Mass
from aviary.subsystems.mass.mass_builder import MassBuilderBase
from aviary.subsystems.mass.mass_builder import CoreMassBuilder

# Propulsion
from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.subsystems.propulsion.engine_model import EngineModel
from aviary.subsystems.propulsion.propulsion_builder import PropulsionBuilderBase
from aviary.subsystems.propulsion.propulsion_builder import CorePropulsionBuilder



"""
Builder for an OpenAeroStruct component that computes a new wing mass.

"""
import openmdao.api as om
import aviary.api as av
from aviary.examples.external_subsystems.OAS_weight.OAS_wing_weight_analysis import OAStructures


class OASWingWeightBuilder(av.SubsystemBuilderBase):
    def __init__(self, name='wing_weight'):
        super().__init__(name)

    def build_pre_mission(self, aviary_inputs):
        '''
        Build an OpenMDAO system for the pre-mission computations of the subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission (formerly statics) part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        '''

        wing_group = om.Group()
        wing_group.add_subsystem("aerostructures",
                                 OAStructures(
                                     symmetry=True,
                                     wing_weight_ratio=1.0,
                                     S_ref_type='projected',
                                     n_point_masses=1,
                                     num_twist_cp=4,
                                     num_box_cp=51),
                                 promotes_inputs=[
                                     ('fuel', av.Mission.Design.FUEL_MASS),
                                 ],
                                 promotes_outputs=[('wing_weight', av.Aircraft.Wing.MASS)])

        return wing_group



from aviary.subsystems.propulsion.propulsion_builder import CorePropulsionBuilder
from aviary.subsystems.geometry.geometry_builder import CoreGeometryBuilder
from aviary.subsystems.mass.mass_builder import CoreMassBuilder
from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.variable_info.variable_meta_data import _MetaData as BaseMetaData
from aviary.variable_info.variables import Dynamic, Mission
from aviary.variable_info.enums import LegacyCode

FLOPS = LegacyCode.FLOPS

prop = CorePropulsionBuilder('core_propulsion', BaseMetaData)
mass = CoreMassBuilder('core_mass', BaseMetaData, FLOPS)
aero = CoreAerodynamicsBuilder('core_aerodynamics', BaseMetaData, FLOPS)
geom = CoreGeometryBuilder('core_geometry', BaseMetaData, FLOPS)

default_premission_subsystems = [prop, geom, mass, aero]
default_mission_subsystems = [aero, prop]


phase_info = {
    "pre_mission": {"include_takeoff": False, "optimize_mass": True},
    "climb": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.2, "unitless"),
            "final_mach": (0.72, "unitless"),
            "mach_bounds": ((0.18, 0.74), "unitless"),
            "initial_altitude": (0.0, "ft"),
            "final_altitude": (32000.0, "ft"),
            "altitude_bounds": ((0.0, 34000.0), "ft"),
            "throttle_enforcement": "path_constraint",
            "fix_initial": True,
            "constrain_final": False,
            "fix_duration": False,
            "initial_bounds": ((0.0, 0.0), "min"),
            "duration_bounds": ((64.0, 192.0), "min"),
            "add_initial_mass_constraint": False,
        },
    },
    "cruise": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.72, "unitless"),
            "final_mach": (0.72, "unitless"),
            "mach_bounds": ((0.7, 0.74), "unitless"),
            "initial_altitude": (32000.0, "ft"),
            "final_altitude": (34000.0, "ft"),
            "altitude_bounds": ((23000.0, 38000.0), "ft"),
            "throttle_enforcement": "boundary_constraint",
            "fix_initial": False,
            "constrain_final": False,
            "fix_duration": False,
            "initial_bounds": ((64.0, 192.0), "min"),
            "duration_bounds": ((56.5, 169.5), "min"),
        },
    },
    "descent": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.72, "unitless"),
            "final_mach": (0.36, "unitless"),
            "mach_bounds": ((0.34, 0.74), "unitless"),
            "initial_altitude": (34000.0, "ft"),
            "final_altitude": (500.0, "ft"),
            "altitude_bounds": ((0.0, 38000.0), "ft"),
            "throttle_enforcement": "path_constraint",
            "fix_initial": False,
            "constrain_final": True,
            "fix_duration": False,
            "initial_bounds": ((120.5, 361.5), "min"),
            "duration_bounds": ((29.0, 87.0), "min"),
        },
    },
    "post_mission": {
        "include_landing": False,
        "constrain_range": True,
        "target_range": (1906., "nmi"),
    },
}


def phase_info_parameterization(phase_info, post_mission_info, aviary_inputs):
    """
    Modify the values in the phase_info dictionary to accommodate different values
    for the following mission design inputs: cruise altitude, cruise Mach number,
    cruise range, design gross mass.

    Parameters
    ----------
    phase_info : dict
        Dictionary of phase settings for a mission profile
    aviary_inputs : <AviaryValues>
        Object containing values and units for all aviary inputs and options

    Returns
    -------
    dict
        Modified phase_info that has been changed to match the new mission
        parameters
    """

    alt_cruise = aviary_inputs.get_val(Mission.Design.CRUISE_ALTITUDE, units='ft')
    mach_cruise = aviary_inputs.get_val(Mission.Summary.CRUISE_MACH)

    # Range
    old_range_cruise, range_units = post_mission_info['target_range']
    range_cruise = aviary_inputs.get_val(Mission.Design.RANGE, units=range_units)
    if range_cruise != old_range_cruise:
        new_val = post_mission_info['target_range'][0] * range_cruise / old_range_cruise
        post_mission_info['target_range'] = (new_val, range_units)

    # Altitude
    old_alt_cruise = 32000.
    if alt_cruise != old_alt_cruise:
        phase_info['climb']['user_options']['final_altitude'] = (alt_cruise, 'ft')
        phase_info['cruise']['user_options']['initial_altitude'] = (alt_cruise, 'ft')
        phase_info['cruise']['user_options']['final_altitude'] = (alt_cruise, 'ft')
        phase_info['descent']['user_options']['initial_altitude'] = (alt_cruise, 'ft')

    # Mach
    old_mach_cruise = 0.72
    if mach_cruise != old_mach_cruise:
        phase_info['climb']['user_options']['final_mach'] = (mach_cruise, "unitless")
        phase_info['cruise']['user_options']['initial_mach'] = (mach_cruise, "unitless")
        phase_info['cruise']['user_options']['final_mach'] = (mach_cruise, "unitless")
        phase_info['descent']['user_options']['initial_mach'] = (mach_cruise, "unitless")

    return phase_info, post_mission_info



import unittest
from openmdao.utils.testing_utils import use_tempdirs

from aviary.utils.process_input_decks import create_vehicle, parse_inputs
from aviary.utils.functions import get_path


# @use_tempdirs
class TestCreateVehicle(unittest.TestCase):

    def test_load_aircraft_csv(self):
        """Test loading a standard aircraft CSV file."""
        file_path = 'models/test_aircraft/aircraft_for_bench_FwFm.csv'
        aircraft_values, initial_guesses = create_vehicle(get_path(file_path))
        self.assertIsNotNone(aircraft_values)
        self.assertIsNotNone(initial_guesses)

    def test_load_modified_aircraft_csv(self):
        """Test loading a modified aircraft CSV file with an additional blank line."""
        original_file_path = 'models/test_aircraft/aircraft_for_bench_FwFm.csv'
        modified_file_path = 'modified_aircraft.csv'

        # Copy and modify the original CSV
        with open(get_path(original_file_path), 'r') as original_file, open(modified_file_path, 'w') as modified_file:
            content = original_file.readlines()
            half_way_point = len(content) // 2

            # Write first half of the file
            modified_file.writelines(content[:half_way_point])

            # Insert a few blank lines
            modified_file.write(',,,,\n')

            # Write second half of the file
            modified_file.writelines(content[half_way_point:])

        # Test create_vehicle with the modified file
        aircraft_values, initial_guesses = create_vehicle(modified_file_path)
        self.assertIsNotNone(aircraft_values)
        self.assertIsNotNone(initial_guesses)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from openmdao.api import ExplicitComponent

from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft, Mission
from aviary.examples.external_subsystems.battery.battery_variable_meta_data import ExtendedMetaData


class CellComp(ExplicitComponent):
    """
    Compute behavior of a single battery cell then expand to the full pack.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        n = self.options['num_nodes']

        # Inputs
        add_aviary_input(self, Mission.Battery.CURRENT, val=3.25 *
                         np.ones(n), meta_data=ExtendedMetaData)
        # Static Constants
        add_aviary_input(self, Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX,
                         meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.N_PARALLEL,
                         val=40., meta_data=ExtendedMetaData)
        add_aviary_input(self, Aircraft.Battery.N_SERIES,
                         val=128.0, meta_data=ExtendedMetaData)
        # Integrated State Variables
        add_aviary_input(self, Mission.Battery.STATE_OF_CHARGE,
                         val=0.98*np.ones(n), meta_data=ExtendedMetaData)
        add_aviary_input(self, Mission.Battery.VOLTAGE_THEVENIN,
                         val=np.ones(n), meta_data=ExtendedMetaData)
        # Map Inputs From The Interpolation Component
        self.add_input('U_oc', val=4.16*np.ones(n),
                       units='V', desc='Open-circuit voltage')
        self.add_input('C_Th', val=2000.*np.ones(n), units='F',
                       desc='Thevenin RC parallel capacitance (polarization)')
        self.add_input('R_Th', val=0.01*np.ones(n), units='ohm',
                       desc='Thevenin RC parallel resistance (polarization)')
        self.add_input('R_0', val=0.01*np.ones(n), units='ohm',
                       desc='Internal resistance of the battery')

        # Outputs
        add_aviary_output(self, Mission.Battery.STATE_OF_CHARGE_RATE,
                          val=np.ones(n), meta_data=ExtendedMetaData)
        add_aviary_output(self, Mission.Battery.VOLTAGE_THEVENIN_RATE,
                          val=np.ones(n), meta_data=ExtendedMetaData)
        add_aviary_output(self, Mission.Battery.VOLTAGE, val=416 *
                          np.ones(n), meta_data=ExtendedMetaData)
        add_aviary_output(self, Mission.Battery.HEAT_OUT,
                          val=np.ones(n), meta_data=ExtendedMetaData)
        add_aviary_output(self, Aircraft.Battery.EFFICIENCY,
                          val=np.ones(n), meta_data=ExtendedMetaData)

        # Partials
        self.declare_partials(of='*', wrt='*', dependent=False)
        ar = np.arange(n)

        self.declare_partials(of=Mission.Battery.STATE_OF_CHARGE_RATE,
                              wrt=Mission.Battery.CURRENT, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.STATE_OF_CHARGE_RATE,
                              wrt=Aircraft.Battery.N_PARALLEL)
        self.declare_partials(of=Mission.Battery.STATE_OF_CHARGE_RATE,
                              wrt=Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX)

        self.declare_partials(of=Mission.Battery.VOLTAGE_THEVENIN_RATE,
                              wrt=Mission.Battery.VOLTAGE_THEVENIN, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE_THEVENIN_RATE,
                              wrt='R_Th', rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE_THEVENIN_RATE,
                              wrt='C_Th', rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE_THEVENIN_RATE,
                              wrt=Mission.Battery.CURRENT, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE_THEVENIN_RATE,
                              wrt=Aircraft.Battery.N_PARALLEL)

        self.declare_partials(of=Mission.Battery.VOLTAGE, wrt='U_oc', rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE,
                              wrt=Mission.Battery.VOLTAGE_THEVENIN, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE,
                              wrt=Mission.Battery.CURRENT, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE,
                              wrt=Aircraft.Battery.N_PARALLEL)
        self.declare_partials(of=Mission.Battery.VOLTAGE, wrt='R_0', rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.VOLTAGE,
                              wrt=Aircraft.Battery.N_SERIES)

        self.declare_partials(of=Mission.Battery.HEAT_OUT,
                              wrt=Mission.Battery.CURRENT, rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.HEAT_OUT,
                              wrt=Aircraft.Battery.N_PARALLEL)
        self.declare_partials(of=Mission.Battery.HEAT_OUT,
                              wrt=Aircraft.Battery.N_SERIES)
        self.declare_partials(of=Mission.Battery.HEAT_OUT, wrt='R_0', rows=ar, cols=ar)
        self.declare_partials(of=Mission.Battery.HEAT_OUT,
                              wrt=Mission.Battery.VOLTAGE_THEVENIN, rows=ar, cols=ar)

        self.declare_partials(of=Aircraft.Battery.EFFICIENCY,
                              wrt=Mission.Battery.CURRENT, rows=ar, cols=ar)
        self.declare_partials(of=Aircraft.Battery.EFFICIENCY,
                              wrt=Aircraft.Battery.N_PARALLEL)
        # self.declare_partials(of=Aircraft.Battery.EFFICIENCY, wrt=Aircraft.Battery.N_SERIES, rows=ar, cols=ar) # known 0
        self.declare_partials(of=Aircraft.Battery.EFFICIENCY,
                              wrt='R_0', rows=ar, cols=ar)
        self.declare_partials(of=Aircraft.Battery.EFFICIENCY,
                              wrt=Mission.Battery.VOLTAGE_THEVENIN, rows=ar, cols=ar)
        self.declare_partials(of=Aircraft.Battery.EFFICIENCY,
                              wrt='U_oc', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        C_Th = inputs['C_Th']
        R_Th = inputs['R_Th']
        U_Th = inputs[Mission.Battery.VOLTAGE_THEVENIN]
        R_0 = inputs['R_0']
        Q_max = inputs[Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX]
        U_oc = inputs['U_oc']
        n_p = inputs[Aircraft.Battery.N_PARALLEL]
        n_s = inputs[Aircraft.Battery.N_SERIES]
        I_pack = inputs[Mission.Battery.CURRENT]

        # cell
        I_Li = I_pack / n_p
        outputs[Mission.Battery.STATE_OF_CHARGE_RATE] = -I_Li / \
            (3600.0 * Q_max)  # conversion from hours to seconds
        outputs[Mission.Battery.VOLTAGE_THEVENIN_RATE] = - \
            U_Th / (R_Th * C_Th) + I_Li / C_Th
        U_L = U_oc - U_Th - (I_Li * R_0)

        # thermal
        Q_cell = I_Li**2 * (R_0 + U_Th/I_Li)

        # pack
        outputs[Mission.Battery.VOLTAGE] = U_L * n_s
        outputs[Mission.Battery.HEAT_OUT] = Q_cell * n_s * n_p
        P_tot = I_pack * outputs[Mission.Battery.VOLTAGE]
        outputs[Aircraft.Battery.EFFICIENCY] = 1. - outputs[Mission.Battery.HEAT_OUT] / \
            ((P_tot + outputs[Mission.Battery.HEAT_OUT]))

    def compute_partials(self, inputs, partials):

        C_Th = inputs['C_Th']
        R_Th = inputs['R_Th']
        U_Th = inputs[Mission.Battery.VOLTAGE_THEVENIN]
        R_0 = inputs['R_0']
        Q_max = inputs[Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX]
        U_oc = inputs['U_oc']
        n_p = inputs[Aircraft.Battery.N_PARALLEL]
        n_s = inputs[Aircraft.Battery.N_SERIES]
        I_pack = inputs[Mission.Battery.CURRENT]

        I_Li = I_pack / n_p
        U_L = U_oc - U_Th - (I_Li * R_0)
        Q_cell = I_Li**2 * (R_0 + U_Th/I_Li)
        U_pack = U_L * n_s
        Q_pack = Q_cell * n_s * n_p
        P_tot = I_pack * U_pack
        pack_eta = 1. - Q_pack/((P_tot + Q_pack))

        dI_li__dnp = -I_pack/n_p**2
        dU_L__dnp = -R_0*dI_li__dnp
        dU_pack__dnp = dU_L__dnp * n_s
        dQ_cell__dnp = 2*I_Li*dI_li__dnp*(R_0+U_Th/I_Li) - U_Th*dI_li__dnp
        dQ_pack__dnp = n_s*(dQ_cell__dnp*n_p + Q_cell)
        dPtot__dnp = I_pack * dU_pack__dnp
        dpack_eta__dnp = -dQ_pack__dnp / \
            (P_tot+Q_pack) + Q_pack/(P_tot + Q_pack)**2 * (dPtot__dnp+dQ_pack__dnp)

        dI_li__dI_pack = 1/n_p
        dQ_cell__dI_pack = 2*I_Li*dI_li__dI_pack*(R_0+U_Th/I_Li) - U_Th*dI_li__dI_pack
        dQ_pack__dI_pack = dQ_cell__dI_pack * n_s * n_p
        dU_pack__dI_pack = -R_0*n_s/n_p
        dPtot__dI_pack = U_pack + dU_pack__dI_pack*I_pack
        dpack_eta__dI_pack = -dQ_pack__dI_pack / \
            (P_tot+Q_pack) + Q_pack/(P_tot + Q_pack)**2 * (dPtot__dI_pack+dQ_pack__dI_pack)

        dPtot__dU_th = -I_pack*n_s
        dQ_pack_dU_th = I_pack * n_s
        dpack_eta__dU_t = -dQ_pack_dU_th / \
            (P_tot+Q_pack) + Q_pack/(P_tot + Q_pack)**2 * (dPtot__dU_th+dQ_pack_dU_th)

        dU_pack__dR_0 = -I_pack * n_s / n_p
        dPtot__dR_0 = I_pack * dU_pack__dR_0
        dQ_pack__dR_0 = (I_pack**2 * n_s) / n_p
        dpack_eta__dR_0 = -dQ_pack__dR_0 / \
            (P_tot+Q_pack) + Q_pack/(P_tot + Q_pack)**2 * (dPtot__dR_0+dQ_pack__dR_0)

        dPtot__dU_oc = I_pack*n_s
        dpack_eta__dU_oc = Q_pack/(P_tot + Q_pack)**2 * dPtot__dU_oc

        partials[Mission.Battery.STATE_OF_CHARGE_RATE,
                 Mission.Battery.CURRENT] = -1./(3600.0*Q_max*n_p)
        partials[Mission.Battery.STATE_OF_CHARGE_RATE,
                 Aircraft.Battery.N_PARALLEL] = I_pack/(3600.0*Q_max*n_p**2)
        partials[Mission.Battery.STATE_OF_CHARGE_RATE,
                 Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX] = I_pack/(3600.0*n_p*Q_max**2)

        partials[Mission.Battery.VOLTAGE_THEVENIN_RATE,
                 Mission.Battery.VOLTAGE_THEVENIN] = -1./(R_Th*C_Th)
        partials[Mission.Battery.VOLTAGE_THEVENIN_RATE, 'R_Th'] = U_Th/(C_Th * R_Th**2)
        partials[Mission.Battery.VOLTAGE_THEVENIN_RATE, 'C_Th'] = (
            n_p*U_Th - I_pack*R_Th) / (C_Th**2 * n_p * R_Th)
        partials[Mission.Battery.VOLTAGE_THEVENIN_RATE,
                 Mission.Battery.CURRENT] = 1./(C_Th*n_p)
        partials[Mission.Battery.VOLTAGE_THEVENIN_RATE,
                 Aircraft.Battery.N_PARALLEL] = -I_pack/(C_Th*n_p**2)

        partials[Mission.Battery.VOLTAGE, 'U_oc'] = n_s
        partials[Mission.Battery.VOLTAGE, Mission.Battery.VOLTAGE_THEVENIN] = -n_s
        partials[Mission.Battery.VOLTAGE, 'R_0'] = dU_pack__dR_0
        partials[Mission.Battery.VOLTAGE, Aircraft.Battery.N_PARALLEL] = dU_pack__dnp
        partials[Mission.Battery.VOLTAGE, Mission.Battery.CURRENT] = dU_pack__dI_pack
        partials[Mission.Battery.VOLTAGE, Aircraft.Battery.N_SERIES] = U_oc - \
            U_Th - (I_pack/n_p * R_0)

        partials[Mission.Battery.HEAT_OUT, Mission.Battery.CURRENT] = n_s * \
            ((2 * I_pack * R_0)/n_p + U_Th)
        partials[Mission.Battery.HEAT_OUT, Aircraft.Battery.N_PARALLEL] = dQ_pack__dnp
        partials[Mission.Battery.HEAT_OUT, 'R_0'] = (I_pack**2 * n_s) / n_p
        partials[Mission.Battery.HEAT_OUT,
                 Mission.Battery.VOLTAGE_THEVENIN] = dQ_pack_dU_th
        partials[Mission.Battery.HEAT_OUT, Aircraft.Battery.N_SERIES] = (
            I_pack**2 * ((n_p * U_Th)/I_pack + R_0)) / n_p

        partials[Aircraft.Battery.EFFICIENCY, 'U_oc'] = dpack_eta__dU_oc
        partials[Aircraft.Battery.EFFICIENCY,
                 Aircraft.Battery.N_PARALLEL] = dpack_eta__dnp
        partials[Aircraft.Battery.EFFICIENCY,
                 Mission.Battery.CURRENT] = dpack_eta__dI_pack
        partials[Aircraft.Battery.EFFICIENCY, 'R_0'] = dpack_eta__dR_0
        partials[Aircraft.Battery.EFFICIENCY,
                 Mission.Battery.VOLTAGE_THEVENIN] = dpack_eta__dU_t
        # partials[Aircraft.Battery.EFFICIENCY,Aircraft.Battery.N_SERIES] = 0.


if __name__ == "__main__":

    from openmdao.api import Problem, Group, IndepVarComp

    p = Problem()
    p.model = Group()

    p.model.add_subsystem('CellComp', CellComp(num_nodes=1), promotes=['*'])

    p.setup(mode='auto', check=True, force_alloc_complex=True)

    p.check_partials(compact_print=True, method='cs', step=1e-50)



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary import constants
from aviary.mission.flops_based.phases.simplified_landing import (LandingCalc,
                                                                  LandingGroup)
from aviary.variable_info.variables import Aircraft, Mission


class LandingCalcTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "land",
            LandingCalc(),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Landing.TOUCHDOWN_MASS, val=152800.0, units="lbm"
        )  # check (this is the design landing mass)
        self.prob.model.set_input_defaults(
            "rho", val=constants.RHO_SEA_LEVEL_METRIC, units="kg/m**3"
        )  # not exact value but should be close enough
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.0, units="ft**2"
        )  # check (this is the reference wing area)
        self.prob.model.set_input_defaults(
            Mission.Landing.LIFT_COEFFICIENT_MAX, val=3, units='unitless')  # check

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()

        tol = 1e-5

        assert_near_equal(
            self.prob[Mission.Landing.GROUND_DISTANCE], 6403.64963504, tol
        )  # not actual value
        # not actual value
        assert_near_equal(
            self.prob[Mission.Landing.INITIAL_VELOCITY], 136.22914933, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class LandingGroupTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "land",
            LandingGroup(),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Mission.Landing.TOUCHDOWN_MASS, val=152800.0, units="lbm"
        )  # check (this is the design landing mass)
        self.prob.model.set_input_defaults(
            Mission.Landing.INITIAL_ALTITUDE, val=35, units="ft"
        )  # confirm initial altitude should be 35 ft.
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.0, units="ft**2"
        )  # check (this is the reference wing area)
        self.prob.model.set_input_defaults(
            Mission.Landing.LIFT_COEFFICIENT_MAX, val=3, units='unitless')  # check

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        self.prob.run_model()

        tol = 1e-5

        assert_near_equal(
            self.prob[Mission.Landing.GROUND_DISTANCE], 6407.19354429, tol
        )  # not actual value
        # not actual value
        assert_near_equal(
            self.prob[Mission.Landing.INITIAL_VELOCITY], 136.22914933, tol)

        partial_data = self.prob.check_partials(
            out_stream=None, excludes=['*.USatm'], method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



'''
Define the ODEs for landing.

Classes
-------
FlareODE : the ODE for the flare phase of landing
'''
import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.flops_based.ode.landing_eom import FlareEOM, StallSpeed
from aviary.mission.flops_based.ode.takeoff_ode import TakeoffODE as _TakeoffODE
from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.functions import set_aviary_initial_values, promote_aircraft_and_mission_vars
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.variable_info.variables_in import VariablesIn


class ExternalSubsystemGroup(om.Group):
    def configure(self):
        promote_aircraft_and_mission_vars(self)


class LandingODE(_TakeoffODE):
    '''
    Define the ODE for most phases of landing.
    '''
    # region : derived type customization points
    stall_speed_lift_coefficient_name = Mission.Landing.LIFT_COEFFICIENT_MAX
    # endregion : derived type customization points


class FlareODE(om.Group):
    '''
    Define the ODE for the flare phase of landing.
    '''

    def initialize(self):
        options = self.options

        options.declare(
            'num_nodes', default=1, types=int,
            desc='Number of nodes to be evaluated in the RHS')

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

        self.options.declare(
            'subsystem_options', types=dict, default={},
            desc='dictionary of parameters to be passed to the subsystem builders')

        self.options.declare(
            'core_subsystems',
            desc='list of core subsystem builder instances to be added to the ODE'
        )
        self.options.declare(
            'external_subsystems', default=[],
            desc='list of external subsystem builder instances to be added to the ODE')

    def setup(self):
        options = self.options

        nn = options["num_nodes"]
        aviary_options = options['aviary_options']
        subsystem_options = options['subsystem_options']
        core_subsystems = options['core_subsystems']

        self.add_subsystem(
            'input_port', VariablesIn(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(num_nodes=nn),
            promotes_inputs=[("h", Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho", ("sos", Dynamic.Mission.SPEED_OF_SOUND), ("temp",
                                                                 Dynamic.Mission.TEMPERATURE),
                ("pres", Dynamic.Mission.STATIC_PRESSURE), "viscosity"])

        self.add_subsystem(
            "fc",
            FlightConditions(num_nodes=nn),
            promotes_inputs=["rho", Dynamic.Mission.SPEED_OF_SOUND,
                             ("TAS", Dynamic.Mission.VELOCITY)],
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.MACH, "EAS"])

        # NOTE: the following are potentially signficant differences in implementation
        # between FLOPS and Aviary:
        #    - FLOPS detailed takeoff/landing assumes constant mass for the duration of
        #      that specific analysis.
        #    - Aviary implementation of FLOPS based detailed takeoff/landing will allow
        #      mass to vary as needed as a function of time and variation in related
        #      optimization control variables.
        self.add_subsystem(
            "stall_speed", StallSpeed(num_nodes=nn),
            promotes_inputs=[
                "mass",
                ("density", "rho"),
                ('area', Aircraft.Wing.AREA),
                ("lift_coefficient_max", Mission.Landing.LIFT_COEFFICIENT_MAX),
            ],
            promotes_outputs=[("stall_speed", "v_stall")])

        base_options = {'num_nodes': nn, 'aviary_inputs': aviary_options}

        for subsystem in core_subsystems:
            # check if subsystem_options has entry for a subsystem of this name
            if subsystem.name in subsystem_options:
                kwargs = subsystem_options[subsystem.name]
            else:
                kwargs = {}

            kwargs.update(base_options)
            system = subsystem.build_mission(**kwargs)

            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        # Create a lightly modified version of an OM group to add external subsystems
        # to the ODE with a special configure() method that promotes
        # all aircraft:* and mission:* variables to the ODE.
        external_subsystem_group = ExternalSubsystemGroup()
        add_subsystem_group = False

        for subsystem in self.options['external_subsystems']:
            subsystem_mission = subsystem.build_mission(
                num_nodes=nn, aviary_inputs=aviary_options)
            if subsystem_mission is not None:
                add_subsystem_group = True
                external_subsystem_group.add_subsystem(subsystem.name, subsystem_mission)

        # Only add the external subsystem group if it has at least one subsystem.
        # Without this logic there'd be an empty OM group added to the ODE.
        if add_subsystem_group:
            self.add_subsystem(
                name='external_subsystems',
                subsys=external_subsystem_group,
                promotes_inputs=['*'],
                promotes_outputs=['*'])

        kwargs = {
            'num_nodes': nn,
            'aviary_options':  options['aviary_options']}

        self.add_subsystem(
            'eoms',
            FlareEOM(**kwargs),
            promotes_inputs=[
                Dynamic.Mission.FLIGHT_PATH_ANGLE, Dynamic.Mission.VELOCITY, Dynamic.Mission.MASS, Dynamic.Mission.LIFT,
                Dynamic.Mission.THRUST_TOTAL, Dynamic.Mission.DRAG, 'angle_of_attack',
                'angle_of_attack_rate', Mission.Landing.FLARE_RATE
            ],
            promotes_outputs=[
                Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE, 'forces_perpendicular',
                'required_thrust', 'net_alpha_rate'
            ]
        )

        self.add_subsystem(
            'comp_v_ratio',
            om.ExecComp(
                'v_over_v_stall = v / v_stall',
                v_over_v_stall={'units': 'unitless', 'shape': nn},
                v={'units': 'm/s', 'shape': nn},
                # NOTE: FLOPS detailed takeoff stall speed is not dynamic - see above
                v_stall={'units': 'm/s', 'shape': nn}),
            promotes_inputs=[('v', Dynamic.Mission.VELOCITY), 'v_stall'],
            promotes_outputs=['v_over_v_stall'])

        self.set_input_defaults(Dynamic.Mission.ALTITUDE, np.zeros(nn), 'm')
        self.set_input_defaults(Dynamic.Mission.VELOCITY, np.zeros(nn), 'm/s')

        set_aviary_initial_values(self, aviary_options)



import numpy as np
import openmdao.api as om
from aviary.constants import GRAV_METRIC_FLOPS as gravity
from aviary.variable_info.variables import Dynamic


class SpecificEnergyRate(om.ExplicitComponent):
    """
    Rutowski "Energy Approach to the General Aircraft Performance Problem", doi 10.2514/8.2956
    Equation 5
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(
            Dynamic.Mission.VELOCITY,
            val=np.ones(nn),
            desc='current velocity',
            units='m/s')
        self.add_input(
            Dynamic.Mission.MASS,
            val=np.ones(nn),
            desc='current mass',
            units='kg')
        self.add_input(Dynamic.Mission.THRUST_TOTAL, val=np.ones(nn),
                       desc='current thrust', units='N')
        self.add_input(
            Dynamic.Mission.DRAG,
            val=np.ones(nn),
            desc='current drag',
            units='N')
        self.add_output(Dynamic.Mission.SPECIFIC_ENERGY_RATE, val=np.ones(
            nn), desc='current specific power', units='m/s')

    def compute(self, inputs, outputs):
        velocity = inputs[Dynamic.Mission.VELOCITY]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]
        weight = inputs[Dynamic.Mission.MASS] * gravity
        outputs[Dynamic.Mission.SPECIFIC_ENERGY_RATE] = velocity * \
            (thrust - drag) / weight

    def setup_partials(self):
        arange = np.arange(self.options['num_nodes'])
        self.declare_partials(Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                              [Dynamic.Mission.VELOCITY,
                               Dynamic.Mission.MASS,
                               Dynamic.Mission.THRUST_TOTAL,
                               Dynamic.Mission.DRAG],
                              rows=arange,
                              cols=arange)

    def compute_partials(self, inputs, J):
        velocity = inputs[Dynamic.Mission.VELOCITY]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        drag = inputs[Dynamic.Mission.DRAG]
        weight = inputs[Dynamic.Mission.MASS] * gravity

        J[Dynamic.Mission.SPECIFIC_ENERGY_RATE,
          Dynamic.Mission.VELOCITY] = (thrust - drag) / weight
        J[Dynamic.Mission.SPECIFIC_ENERGY_RATE,
          Dynamic.Mission.THRUST_TOTAL] = velocity / weight
        J[Dynamic.Mission.SPECIFIC_ENERGY_RATE, Dynamic.Mission.DRAG] = -velocity / weight
        J[Dynamic.Mission.SPECIFIC_ENERGY_RATE, Dynamic.Mission.MASS] = -gravity\
            * velocity * (thrust - drag) / (weight)**2



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.cargo import CargoMass
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import do_validation_test, print_case
from aviary.variable_info.variables import Aircraft

cargo_test_data = {}
cargo_test_data['1'] = AviaryValues({
    Aircraft.CrewPayload.BAGGAGE_MASS_PER_PASSENGER: (50, 'lbm'),
    Aircraft.CrewPayload.MISC_CARGO: (2000., 'lbm'),  # custom
    Aircraft.CrewPayload.MASS_PER_PASSENGER: (180., 'lbm'),
    Aircraft.CrewPayload.WING_CARGO: (1000., 'lbm'),  # custom
    Aircraft.CrewPayload.BAGGAGE_MASS: (9200., 'lbm'),  # custom
    Aircraft.CrewPayload.NUM_PASSENGERS: (184, 'unitless'),  # custom
    Aircraft.CrewPayload.PASSENGER_MASS: (33120., 'lbm'),  # custom
    Aircraft.CrewPayload.CARGO_MASS: (3000., 'lbm'),  # custom
    Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS: (45320., 'lbm')  # custom
})

cargo_data_sets = [key for key in cargo_test_data]


class CargoMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(cargo_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):
        validation_data = cargo_test_data[case_name]
        prob = self.prob

        prob.model.add_subsystem(
            'cargo_passenger',
            CargoMass(aviary_options=validation_data),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=validation_data,
                           output_validation_data=validation_data,
                           input_keys=[Aircraft.CrewPayload.MISC_CARGO,
                                       Aircraft.CrewPayload.WING_CARGO],
                           output_keys=[Aircraft.CrewPayload.BAGGAGE_MASS,
                                        Aircraft.CrewPayload.PASSENGER_MASS,
                                        Aircraft.CrewPayload.CARGO_MASS,
                                        Aircraft.CrewPayload.TOTAL_PAYLOAD_MASS,
                                        ])

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



thickness_to_chord_scaler = 0.387
_fuselage_adjustment_scaler = 0.6730


def calc_lifting_surface_scaler(thickness_to_chord):
    scaler = thickness_to_chord_scaler * thickness_to_chord + 2.0

    return scaler


def calc_fuselage_adjustment(chord, thickness_to_chord):
    adjustment = \
        _fuselage_adjustment_scaler * chord * (thickness_to_chord * chord)

    return adjustment


def d_calc_fuselage_adjustment(chord, thickness_to_chord):
    d1 = 2.0 * _fuselage_adjustment_scaler * thickness_to_chord * chord
    d2 = _fuselage_adjustment_scaler * chord ** 2

    return d1, d2


class Names:
    '''
    Define component I/O variable names that should not exported.
    '''
    CROOT = 'prep_geom:_Names:CROOT'
    CROOTB = 'prep_geom:_Names:CROOTB'
    CROTM = 'prep_geom:_Names:CROTM'
    CROTVT = 'prep_geom:_Names:CROTVT'
    CRTHTB = 'prep_geom:_Names:CRTHTB'
    # equivalent(?)
    SPANHT = 'prep_geom:_Names:SPANHT'
    SPANVT = 'prep_geom:_Names:SPANVT'
    XDX = 'prep_geom:_Names:XDX'
    XMULT = 'prep_geom:_Names:XMULT'
    XMULTH = 'prep_geom:_Names:XMULTH'
    XMULTV = 'prep_geom:_Names:XMULTV'

    __slots__ = ()



"""
Builder for a simple component that computes a new wing and horizontal tail
mass.

The SimpleWeight component is placed inside of a group so that we can promote
the variable "Tail" using the alias Aircraft.HorizontalTail.MASS
"""
import openmdao.api as om

from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.variable_info.variables import Aircraft

from aviary.examples.external_subsystems.simple_weight.simple_weight import SimpleWeight


class WingWeightBuilder(SubsystemBuilderBase):
    """
    Prototype of a subsystem that overrides an aviary internally computed var.
    """

    def __init__(self, name='wing_weight'):
        super().__init__(name)

    def build_pre_mission(self, aviary_inputs):
        '''
        Build an OpenMDAO system for the pre-mission computations of the subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission (formerly statics) part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        '''
        wing_group = om.Group()
        wing_group.add_subsystem("aerostructures", SimpleWeight(),
                                 promotes_inputs=["aircraft:*"],
                                 promotes_outputs=[Aircraft.Wing.MASS,
                                                   ('Tail', Aircraft.HorizontalTail.MASS)])
        return wing_group



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.aerodynamics.flops_based.mach_number import MachNumber
from aviary.variable_info.variables import Dynamic


class MachNumberTest(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            Dynamic.Mission.MACH,
            MachNumber(num_nodes=1),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):
        # for key, temp in FLOPS_Test_Data.items():
        # TODO currently no way to use FLOPS test case data for mission components

        self.prob.set_val(Dynamic.Mission.VELOCITY, val=347, units='ft/s')
        self.prob.set_val(Dynamic.Mission.SPEED_OF_SOUND, val=1045, units='ft/s')
        self.prob.run_model()

        tol = 1e-3
        assert_near_equal(
            self.prob.get_val(Dynamic.Mission.MACH, units='unitless'), 0.332, tol
        )  # check the value of each output

        # TODO resolve partials wrt gravity (decide on implementation of gravity)
        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(
            partial_data, atol=1e-6, rtol=1e-6
        )  # check the partial derivatives

    # TODO IO test will fail until mission variable hirerarchy implemented
    # def test_IO(self):
    #     assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



'''
Define utilities for using aviary values with associated units and testing
for compatibility with aviary metadata dictionary.

Utilities
---------
Units : type alias
    define a type hint for associated units

ValueAndUnits : type alias
    define a type hint for a single value paired with its associated units

OptionalValueAndUnits : type alias
    define a type hint for an optional single value paired with its associated units

class AviaryValues
    define a collection of named values with associated units
'''
from enum import EnumMeta

import numpy as np
from openmdao.utils.units import convert_units as _convert_units

from aviary.utils.named_values import (NamedValues, get_items, get_keys,
                                       get_values)
from aviary.variable_info.variable_meta_data import _MetaData


class AviaryValues(NamedValues):
    '''
    Define a collection of aviary values with associated units and aviary tests.
    '''

    def set_val(self, key, val, units='unitless', meta_data=_MetaData):
        '''
        Update the named value and its associated units.

        Note, specifying units of `None` or units of any type other than `str` will raise
        `Typerror`.

        Parameters
        ----------
        key : str
            the name of the item

        val : Any
            the new value of the item

        units : str ('unitless')
            the units associated with the new value, if any

        Raises
        ------
        TypeError
            if units of `None` were specified or units of any type other than `str`
        '''

        # Special handling to access an Enum member from either the member name or its value.
        my_val = val
        if key in _MetaData.keys():
            expected_types = _MetaData[key]['types']
            if type(expected_types) is EnumMeta:
                if self._is_iterable(val):
                    my_val = [self._convert_to_enum(
                        item, expected_types) for item in val]
                else:
                    my_val = self._convert_to_enum(val, expected_types)

        self._check_type(key, my_val, meta_data=meta_data)
        self._check_units_compatability(key, my_val, units, meta_data=meta_data)

        super().set_val(key=key, val=my_val, units=units)

    def _check_type(self, key, val, meta_data=_MetaData):
        if key in meta_data.keys():
            expected_types = meta_data[key]['types']
            if expected_types is not None:
                if isinstance(expected_types, list):
                    expected_types = tuple(expected_types)
                # if val is not iterable, add it to a list (length 1)
                if not self._is_iterable(val):
                    val = [val]
                # numpy arrays have special typings. Extract item of equivalent built-in python type
                # numpy arrays do not allow mixed types, only have to check first entry
                # empty arrays do not need this check
                if isinstance(val, np.ndarray) and len(val) > 0:
                    # NoneType numpy arrays do not need to be "converted" to built-in python types
                    if val.dtype == type(None):
                        val = [val[0]]
                    else:
                        val = [val[0].item()]
                for item in val:
                    has_bool = False  # needs some fancy shenanigans because bools will register as ints
                    if (isinstance(expected_types, type)):
                        if expected_types is bool:
                            has_bool = True
                    elif bool in expected_types:
                        has_bool = True
                    if (not isinstance(item, expected_types)) or (
                            (has_bool == False) and (isinstance(item, bool))):
                        raise TypeError(
                            f'{key} is of type(s) {meta_data[key]["types"]} but you have provided a value of type {type(item)}.')

    def _check_units_compatability(self, key, val, units, meta_data=_MetaData):
        if key in meta_data.keys():
            expected_units = meta_data[key]['units']

            try:
                # NOTE the value here is unimportant, we only care if OpenMDAO will
                # convert the units
                _convert_units(10, expected_units, units)
            except ValueError:
                raise ValueError(
                    f'The units {units} which you have provided for {key} are invalid.')
            except TypeError:
                raise TypeError(
                    f'The base units of {key} are {expected_units}, and you have tried to set {key} with units of {units}, which are not compatible.')
            except BaseException:
                raise KeyError('There is an unknown error with your units.')

    def _is_iterable(self, val):
        return isinstance(val, list) or isinstance(val, np.ndarray) or isinstance(val, tuple)

    def _convert_to_enum(self, val, enum_type):
        if isinstance(val, str):
            try:
                # see if str maps to ENUM value
                return enum_type(val)
            except ValueError:
                # str instead maps to ENUM name
                return enum_type[val.upper()]
        else:
            return enum_type(val)



"""
Battery Performance Modeling on Maxwell X-57
Jeffrey C. Chin, Sydney L. Schnulo, Thomas B. Miller,
Kevin Prokopius, and Justin Gray
http://openmdao.org/pubs/chin_battery_performance_x57_2019.pdf

Thevenin voltage equation based on paper
"Evaluation of Lithium Ion Battery Equivalent Circuit Models
for State of Charge Estimation by an Experimental Approach"
Hongwen H, Rui Xiong, Jinxin Fan

Equivalent Circuit component values derived from
"High Fidelity Electrical Model with Thermal Dependence for
 Characterization and Simulation of High Power Lithium Battery Cells"
Tarun Huria, Massimo Ceraolo, Javier Gazzarri, Robyn Jackey
"""

from openmdao.api import Group
from .cell_comp import CellComp
from .reg_thevenin_interp_group import RegTheveninInterpGroup

from aviary.utils.aviary_values import AviaryValues
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft, Mission


class BatteryMission(Group):
    """Assembly to connect subcomponents of the Thevenin Battery Equivalent
    Circuit Model From Interpolated Performance Maps
    """

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare(
            'aviary_inputs', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options',
            default=None,
        )

    def setup(self):
        n = self.options["num_nodes"]

        self.add_subsystem(
            name="interp_group",
            subsys=RegTheveninInterpGroup(num_nodes=n),
        )

        self.add_subsystem(name="cell", subsys=CellComp(num_nodes=n))

        # Connect internal names
        self.connect("interp_group.U_oc", "cell.U_oc")
        self.connect("interp_group.C_Th", "cell.C_Th")
        self.connect("interp_group.R_Th", "cell.R_Th")
        self.connect("interp_group.R_0", "cell.R_0")

        # Promote interp group inputs to match Aviary variable names
        self.promotes(
            "interp_group",
            inputs=[
                ("T_batt", "mission:battery:temperature"),
                ("SOC", "mission:battery:state_of_charge"),
            ],
        )

        self.promotes(
            "cell",
            inputs=[
                Mission.Battery.VOLTAGE_THEVENIN,
                Mission.Battery.CURRENT,
                Aircraft.Battery.N_SERIES,
                Aircraft.Battery.N_PARALLEL,
                Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX,
            ],
        )

        self.promotes(
            "cell",
            outputs=[
                Mission.Battery.VOLTAGE,
                Mission.Battery.VOLTAGE_THEVENIN_RATE,
                Mission.Battery.STATE_OF_CHARGE_RATE,
                Mission.Battery.HEAT_OUT,
                Aircraft.Battery.EFFICIENCY,
            ],
        )



"""
Define utilities for building engine models.

Classes
-------
EngineModel : the interface for an engine model builder.
"""
import warnings

import numpy as np

from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.utils.aviary_values import AviaryValues


class EngineModel(SubsystemBuilderBase):
    """
    Define the interface for an engine model builder.

    Attributes
    ----------
    name : str ('engine_model')
        object label.
    options : AviaryValues (<empty>)
        inputs and options related to engine model.

    Methods
    -------
    build_pre_mission
    build_mission
    build_post_mission
    get_val
    set_val
    update
    """

    default_name = 'engine_model'

    def __init__(
        self, name: str = None, options: AviaryValues = None, meta_data: dict = None,
    ):
        super().__init__(name, meta_data=meta_data)
        self.options = options.deepcopy()

        # Hybrid throttle is currently the only optional independent variable, requiring
        # this flag so Aviary knows how to handle EngineModels during mission
        self.use_hybrid_throttle = False

        self._preprocess_inputs()

    def _setup(self, **kwargs):
        """
        Perform setup of EngineModel.

        All attributes and values for options in EngineModel are finalized for
        analysis.
        """
        self._preprocess_inputs()

    def build_pre_mission(self, aviary_inputs):
        """
        Build an OpenMDAO system for the pre-mission computations of the engine model,
        such as sizing.

        Optional for engine models.
        Used in propulsion_sizing.py to build the pre-mission propulsion subsystem.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        """
        return None

    def build_mission(self, num_nodes, aviary_inputs):
        """
        Build an OpenMDAO system for the mission computations of the engine model.

        Required for engine models.
        Used in propulsion_mission.py to build the propulsion mission system.

        Returns
        -------
        mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen
            during the mission. This includes time-dependent states that are
            being integrated as well as any other variables that vary during
            the mission.
        """
        raise NotImplementedError('build_mission() is a required method but has not '
                                  f'been implemented in EngineModel <{self.name}>')

    def build_post_mission(self, aviary_inputs):
        """
        Build an OpenMDAO system for the post-mission computations of the engine model.

        Optional for engine models. Currently unused by core Aviary propulsion.

        Returns
        -------
        post_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen
            after the mission. This includes time-dependent states that are
            being integrated as well as any other variables that vary after
            the mission.
        """
        return None

    def _preprocess_inputs(self):
        """
        Raises TypeError if options are not an instance of AviaryValues (or None).

        Creates warning if a provided engine option is vectorized, and only accepts
        the first entry in that vector into self.options.
        """
        options = self.options

        if options is None:
            return  # options are allowed to be empty

        if not isinstance(options, AviaryValues):
            raise TypeError('EngineModel options must be an AviaryValues object')

        for (key, (val, units)) in options:
            # only perform vector check for variables related to engines and nacelles
            if key.startswith('aircraft:engine:') or key.startswith('aircraft:nacelle'):
                if type(val) in (list, np.ndarray, tuple):
                    if self.meta_data[key]['types'] not in (list, np.ndarray, tuple):
                        warnings.warn(
                            f'The value of {key} passed to EngineModel <{self.name}> is '
                            f'type {type(val)}. Only the first entry in this iterable will '
                            'be used.')
                    if isinstance(val, np.ndarray):
                        # "Convert" numpy types to standard Python types. Wrap first
                        # index in numpy array before calling item() to safeguard against
                        # non-standard types, such as objects
                        val = val.ravel()[0]

                        # Can't setval an int with an int64.
                        if isinstance(val, np.integer):
                            val = int(val)

                    else:
                        val = val[0]
                    # update options with single value (instead of vector)
                    options.set_val(key, val, units)
            # Currently assuming that EngineModels might care about non-engine variables,
            # so they are being kept in self.options
            # else:
            #     options.delete(key)

    def update(self, options: AviaryValues, **kwargs):
        """
        Given a new set of AviaryValues, update the engine model and rerun setup.
        """
        self.options = options.deepcopy()

        self._setup(**kwargs)

    def get_val(self, key, units='unitless'):
        """
        Returns desired value from options in specified units.

        Parameters
        -------
        key : str
            Name of requested option.
        units : str
            Unit requested option value should be converted to.

        Returns
        -------
        val
            Value of requested option in desired units.

        """
        return self.options.get_val(key, units)

    def set_val(self, key, val, units='unitless'):
        """
        Updates desired value in options with specified units.

        Parameters
        -------
        key : str
            Name of option whose value will be updated.
        val
            New value for option.
        units : str
            Unit of val.
        """
        self.options.set_val(key, val, units)
